(***********************************************************************)
(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
(*   \VV/  *************************************************************)
(*    //   *      This file is distributed under the terms of the      *)
(*         *       GNU Lesser General Public License Version 2.1       *)
(***********************************************************************)

(* Finite sets library.  
 * Authors: Pierre Letouzey and Jean-Christophe Filliâtre 
 * Institution: LRI, CNRS UMR 8623 - Université Paris Sud
 *              91405 Orsay, France *)

(* $Id: FSetAVL.v,v 1.17 2003/07/08 13:43:36 filliatr Exp $ *)

(** This module implements sets using AVL trees.
    It follows the implementation from Ocaml's standard library. *)

Require FSetInterface.
Require FSetList.

Require ZArith.
Import Z_scope.
Open Scope Z_scope.

Notation "[]" := (nil ?) (at level 1).
Notation "a :: l" := (cons a l) (at level 1, l at next level).

Set Ground Depth 3.

Module Make [X : OrderedType] <: Sdep with Module E := X.

  Module E := X.
  Module ME := MoreOrderedType X.

  Definition elt := X.t.

  (** * Trees *)

  Inductive tree : Set :=
  | Leaf : tree
  | Node : tree -> X.t -> tree -> Z -> tree.

  (** The fourth field of [Node] is the height of the tree *)

  (** * Occurrence in a tree *)

  Inductive In_tree [x:elt] : tree -> Prop :=
  | IsRoot : (l,r:tree)(h:Z)(y:elt)
             (X.eq x y) -> (In_tree x (Node l y r h))
  | InLeft : (l,r:tree)(h:Z)(y:elt)
             (In_tree x l) -> (In_tree x (Node l y r h))
  | InRight : (l,r:tree)(h:Z)(y:elt)
              (In_tree x r) -> (In_tree x (Node l y r h)).

  Hint In_tree := Constructors In_tree.

  (** [In_tree] is compatible with [X.eq] *)

  Lemma eq_In_tree: (s:tree)(x,y:elt)(E.eq x y) -> (In_tree x s) -> (In_tree y s).
  Proof.
    Induction s; Simpl; Intuition.
    Inversion_clear H0.
    Inversion_clear H2; Intuition EAuto.
  Save.

  (** [In_tree] is height-insensitive *)

  Lemma In_height : (h,h':Z)(x,y:elt)(l,r:tree)
    (In_tree y (Node l x r h)) -> (In_tree y (Node l x r h')).
  Proof.
    Inversion 1; Auto.
  Save.
  Hints Resolve In_height.

  (** * Binary search trees *)

  (** [lt_tree x s]: all elements in [s] are smaller than [x] 
      (resp. greater for [gt_tree]) *)

  Definition lt_tree [x:elt; s:tree] := (y:elt)(In_tree y s) -> (X.lt y x).
  Definition gt_tree [x:elt; s:tree] := (y:elt)(In_tree y s) -> (X.lt x y).

  Hints Unfold lt_tree gt_tree.

  (** Results about [lt_tree] and [gt_tree] *)

  Lemma lt_leaf : (x:elt)(lt_tree x Leaf).
  Proof.
    Unfold lt_tree; Intros; Inversion H.
  Save.

  Lemma gt_leaf : (x:elt)(gt_tree x Leaf).
  Proof.
    Unfold gt_tree; Intros; Inversion H.
  Save.

  Lemma lt_tree_node : (x,y:elt)(l,r:tree)(h:Z)
    (lt_tree x l) -> (lt_tree x r) -> (X.lt y x) -> 
    (lt_tree x (Node l y r h)).
  Proof.
    Unfold lt_tree; Intuition.
    Inversion_clear H2; Intuition.
    Apply ME.eq_lt with y; Auto.
  Save.

  Lemma gt_tree_node : (x,y:elt)(l,r:tree)(h:Z)
    (gt_tree x l) -> (gt_tree x r) -> (E.lt x y) -> 
    (gt_tree x (Node l y r h)).
  Proof.
    Unfold gt_tree; Intuition.
    Inversion_clear H2; Intuition.
    Apply ME.lt_eq with y; Auto.
  Save.

  Hints Resolve lt_leaf gt_leaf lt_tree_node gt_tree_node.

  Lemma lt_node_lt : (x,y:elt)(l,r:tree)(h:Z)
     (lt_tree x (Node l y r h)) -> (E.lt y x).
  Proof.
    Intros; Apply H; Auto.
  Save.

  Lemma gt_node_gt : (x,y:elt)(l,r:tree)(h:Z)
     (gt_tree x (Node l y r h)) -> (E.lt x y).
  Proof.
    Intros; Apply H; Auto.
  Save.

  Lemma lt_left : (x,y:elt)(l,r:tree)(h:Z)
     (lt_tree x (Node l y r h)) -> (lt_tree x l).
  Proof.
    Intros; Red; Intros; Apply H; Auto.
  Save.

  Lemma lt_right : (x,y:elt)(l,r:tree)(h:Z)
     (lt_tree x (Node l y r h)) -> (lt_tree x r).
  Proof.
    Intros; Red; Intros; Apply H; Auto.
  Save.

  Lemma gt_left : (x,y:elt)(l,r:tree)(h:Z)
     (gt_tree x (Node l y r h)) -> (gt_tree x l).
  Proof.
    Intros; Red; Intros; Apply H; Auto.
  Save.

  Lemma gt_right : (x,y:elt)(l,r:tree)(h:Z)
     (gt_tree x (Node l y r h)) -> (gt_tree x r).
  Proof.
    Intros; Red; Intros; Apply H; Auto.
  Save.

  Hints Resolve lt_node_lt gt_node_gt
                lt_left lt_right gt_left gt_right.

  Lemma lt_tree_not_in : 
    (x:elt)(t:tree)(lt_tree x t) -> ~(In_tree x t).
  Proof.
    Unfold lt_tree; Intros; Red; Intros.
    Generalize (H x H0); Intro; Absurd (X.lt x x); Auto.
  Save.

  Lemma lt_tree_trans : 
    (x,y:elt)(X.lt x y) -> (t:tree)(lt_tree x t) -> (lt_tree y t).
  Proof.
    Unfold lt_tree; Ground EAuto.
  Save.

  Lemma gt_tree_not_in : 
    (x:elt)(t:tree)(gt_tree x t) -> ~(In_tree x t).
  Proof.
    Unfold gt_tree; Intros; Red; Intros.
    Generalize (H x H0); Intro; Absurd (X.lt x x); Auto.
  Save.

  Lemma gt_tree_trans : 
    (x,y:elt)(X.lt y x) -> (t:tree)(gt_tree x t) -> (gt_tree y t).
  Proof.
    Unfold gt_tree; Ground EAuto.
  Save.

  Hints Resolve lt_tree_not_in lt_tree_trans 
                gt_tree_not_in gt_tree_trans.

  (** [bst t] : [t] is a binary search tree *)

  Inductive bst : tree -> Prop :=
  | BSLeaf : 
      (bst Leaf)
  | BSNode : (x:elt)(l,r:tree)(h:Z)
      (bst l) -> (bst r) ->
      (lt_tree x l) -> (gt_tree x r) ->
      (bst (Node l x r h)).

  Hint bst := Constructors bst.

  (** Results about [bst] *)
 
  Lemma bst_left : (x:elt)(l,r:tree)(h:Z)
    (bst (Node l x r h)) -> (bst l).
  Proof.
    Intros x l r h H; Inversion H; Auto.
  Save.

  Lemma bst_right : (x:elt)(l,r:tree)(h:Z)
    (bst (Node l x r h)) -> (bst r).
  Proof.
    Intros x l r h H; Inversion H; Auto.
  Save.

  Implicits bst_left. Implicits bst_right.
  Hints Resolve bst_left bst_right.

  Lemma bst_height : (h,h':Z)(x:elt)(l,r:tree)
    (bst (Node l x r h)) -> (bst (Node l x r h')).
  Proof.
    Inversion 1; Auto.
  Save.
  Hints Resolve bst_height.

  (** Key fact about binary search trees: rotations preserve the 
      [bst] property *)

  Lemma rotate_left : (x,y:elt)(a,b,c:tree)(h1,h2,h3,h4:Z)
    (bst (Node a x (Node b y c h2) h1)) ->
    (bst (Node (Node a x b h4) y c h3)).
  Proof.
    Intros; Inversion H; Intuition.
    Constructor; Intuition.
    Constructor; EAuto.
    EAuto.
    Apply lt_tree_node; Intuition.
    Apply lt_tree_trans with x; Auto.
    Inversion H5; Auto.
    Inversion H5; Auto.
  Save.

  Lemma rotate_right : (x,y:elt)(a,b,c:tree)(h1,h2,h3,h4:Z)
    (bst (Node (Node a x b h4) y c h3)) ->
    (bst (Node a x (Node b y c h2) h1)).
  Proof.
    Intros; Inversion H; Intuition.
    Constructor; Intuition.
    EAuto.
    Constructor; Auto.
    Inversion H4; Auto.
    Inversion H4; Auto.
    Apply gt_tree_node; Intuition.
    Inversion H4; Auto.
    Apply gt_tree_trans with y; Auto.
    EAuto.
  Save.

  Hints Resolve rotate_left rotate_right.

  (** * AVL trees *)

  (** [avl s] : [s] is a properly balanced AVL tree,
      i.e. for any node the heights of the two children
      differ by at most 2 *)

  Definition height : tree -> Z :=
    [s:tree]Cases s of
            | Leaf => 0
            | (Node _ _ _ h) => h end.

  Definition max [x,y:Z] : Z := 
    if (Z_lt_ge_dec x y) then [_]y else [_]x.

  (** Instead of writing [h = 1 + (max (height l) (height r))] we prefer 
      the following relation [height_of_node l r h] to ease the use of 
      [Omega] *)

  Definition height_of_node [l,r:tree; h:Z] :=  
    ((height l) >= (height r) /\ h = (height l) + 1) \/
    ((height r) >= (height l) /\ h = (height r) + 1).

  Inductive avl : tree -> Prop :=
  | RBLeaf : 
      (avl Leaf)
  | RBNode : (x:elt)(l,r:tree)(h:Z)
      (avl l) -> (avl r) ->
      `-2 <= (height l) - (height r) <= 2` ->
      (height_of_node l r h) -> 
      (avl (Node l x r h)).

  Hint avl := Constructors avl.

 (** Results about [avl] *)

  Lemma avl_left : 
    (x:elt)(l,r:tree)(h:Z)
    (avl (Node l x r h)) -> (avl l).
  Proof.
    Intros x l r h H; Inversion_clear H; Intuition.
  Save.

  Lemma avl_right : 
    (x:elt)(l,r:tree)(h:Z)
    (avl (Node l x r h)) -> (avl l).
  Proof.
    Intros x l r c H; Inversion_clear H; Intuition.
  Save.

  Implicits avl_left. Implicits avl_right.
  Hints Resolve avl_left avl_right.

  Tactic Definition MaxCase x y := 
    Unfold max; Case (Z_lt_ge_dec x y); Simpl.

  Lemma avl_node: (x:elt)(l,r:tree)
     (avl l) -> (avl r) ->
     `-2 <= (height l) - (height r) <= 2` ->
     (avl (Node l x r ((max (height l) (height r)) + 1))).
  Proof.
    Intros; Constructor; Unfold height_of_node; 
    MaxCase '(height l) '(height r); Intuition.
  Save.
  Hints Resolve avl_node.

  (** The [AVL] tactic *)

  Lemma height_non_negative :
    (s:tree)(avl s) -> (height s) >= 0.
  Proof.
    Induction s; Simpl; Intros.
    Omega.
    Inversion_clear H1; Unfold height_of_node in H5; Intuition.
  Save.
  
  Lemma height_equation : 
    (l,r:tree)(x:elt)(h:Z)
    (avl (Node l x r h)) -> 
    `-2 <= (height l) - (height r) <= 2` /\
    (((height l) >= (height r) /\ h = (height l) + 1) \/
     ((height r) >= (height l) /\ h = (height r) + 1)).
  Proof.
    Inversion 1; Intuition.
  Save.

  Implicits height_non_negative. 
  Implicits height_equation.

  (** If [h] is a proof of [avl (Node l x r h)], then tactic
      [AVL h] is adding all information about [h] to the context *)
  Tactic Definition AVL h :=
    (Generalize (height_non_negative h); Try Simpl);
    (Try Generalize (height_equation h)); Intros.

  (** * Sets as AVL trees *)

  (** A set is implement as a record [t], containing a tree, 
      a proof that it is a binary search tree and a proof that it is 
      a properly balanced AVL tree *)

  Record t_ : Set := t_intro {
    the_tree :> tree; 
    is_bst : (bst the_tree);
    is_avl : (avl the_tree) }.
  Definition t := t_.

  (** Projections *)

  Lemma t_is_bst : (s:t)(bst s).
  Proof.
    Destruct s; Auto.
  Save.
  Hints Resolve t_is_bst.

  Lemma t_is_avl : (s:t)(avl s).
  Proof.
    Destruct s; Auto.
  Save.
  Hints Resolve t_is_avl.

 (** * Logical appartness *)

  Definition In : elt -> t -> Prop := [x:elt][s:t](In_tree x s).

  Definition Equal := [s,s'](a:elt)(In a s)<->(In a s').
  Definition Subset := [s,s'](a:elt)(In a s)->(In a s').
  Definition Add := [x:elt;s,s':t](y:elt)(In y s') <-> ((E.eq y x)\/(In y s)).
  Definition Empty := [s](a:elt)~(In a s).
  Definition For_all := [P:elt->Prop; s:t](x:elt)(In x s)->(P x).
  Definition Exists := [P:elt->Prop; s:t](EX x:elt | (In x s)/\(P x)).

  Lemma eq_In: (s:t)(x,y:elt)(E.eq x y) -> (In x s) -> (In y s).
  Proof.
    Unfold In; Destruct s; Simpl; Intuition Clear is_bst0 is_avl0.
    Induction the_tree0; Inversion_clear H0; Intuition.
    Apply IsRoot; EAuto.
  Save.

  Hints Resolve eq_In.

  (** * Empty set *)

  Definition t_empty : t.
  Proof.
    Exists Leaf; Auto.
  Defined.

  Definition empty : { s:t | (x:elt)~(In x s) }. 
  Proof.
    Exists t_empty.
    Unfold In; Red; Intros.
    Inversion H.
  Defined.

  (** * Emptyness test *)

  Definition is_empty : (s:t){ Empty s }+{ ~(Empty s) }.
  Proof.
    Unfold Empty In; Destruct s; Destruct the_tree0; Simpl; Intros.
    Left; Auto.
    Right; Intuition.
    Apply (H t1); Auto.
  Defined.

  (** * Appartness *)

  (** The [mem] function is deciding appartness. It exploits the [bst] property
      to achieve logarithmic complexity. *)

  Definition mem : (x:elt) (s:t) { (In x s) } + { ~(In x s) }.
  Proof.
    Intros x (s,Hs,Ha).
    Unfold In; Simpl; Clear Ha.
    Generalize Hs; Elim s; Simpl; Intros.
  (* Leaf *)
    Right. 
    Unfold In; Red; Intros; Inversion H.
  (* Node *)
    Elim (X.compare x t1); Intro.
    (* lt x t1 *)
    Case H; Intros.
    EAuto.
    Left; Auto.
    Right; Intro.
    Inversion H1; Intuition.
    Absurd (X.eq x t1); Auto.
    Inversion Hs0.
    Absurd (In_tree x t2); EAuto.
    (* eq x t1 *)
    Left; Auto.
    (* lt t1 x *)
    Case H0; Intros.
    EAuto.
    Left; Auto.
    Right; Intro.
    Inversion H1; Intuition.
    Absurd (X.eq t1 x); Auto.
    Inversion Hs0.
    Absurd (In_tree x t0); EAuto.
  Defined.

  (** * Singleton set *)

  Definition singleton_tree [x:elt] := (Node Leaf x Leaf 1).

  Lemma singleton_bst : (x:elt)(bst (singleton_tree x)).
  Proof.
    Unfold singleton_tree; Auto.
  Save.

  Lemma singleton_avl : (x:elt)(avl (singleton_tree x)).
  Proof.
    Unfold singleton_tree; Intro.
    Constructor; Auto; Unfold height_of_node height; Simpl; Omega.
  Save.

  Definition singleton : (x:elt){ s:t | (y:elt)(In y s) <-> (E.eq x y)}.
  Proof.
    Intro x; Exists (t_intro (singleton_tree x) (singleton_bst x)
             (singleton_avl x)).
    Unfold In singleton_tree; Simpl; Intuition.
    Inversion_clear H; Auto; Inversion H0.
  Defined.

  (** * Helper functions *)

  (** [create l x r] creates a node, assuming [l] and [r]
      to be balanced and [|height l - height r| <= 2]. *)

  Definition create :
    (l:tree)(x:elt)(r:tree)
    (bst l) -> (avl l) -> (bst r) -> (avl r) ->
    (lt_tree x l) -> (gt_tree x r) ->
    `-2 <= (height l) - (height r) <= 2` ->
    { s:tree | 
        (bst s) /\
        (avl s) /\
        (height_of_node l r (height s)) /\
        (y:elt)(In_tree y s) <-> 
            ((X.eq y x) \/ (In_tree y l) \/ (In_tree y r)) }.
  Proof.
    Unfold height_of_node; Intros.
    Exists (Node l x r ((max (height l) (height r)) + 1)).
    Intuition.
    MaxCase '(height l) '(height r); Intuition.
    Inversion_clear H5; Intuition.
  Defined.

  (** [bal l x r] acts as [create], but performs one step of
      rebalancing if necessary, i.e. assumes [|height l - height r| <= 3]. *)

  Definition bal :
    (l:tree)(x:elt)(r:tree)
    (bst l) -> (avl l) -> (bst r) -> (avl r) ->
    (lt_tree x l) -> (gt_tree x r) ->
    `-3 <= (height l) - (height r) <= 3` ->
    { s:tree | 
       (bst s) /\ (avl s) /\
       (* height may be decreased by 1 *)
       (((height_of_node l r (height s)) \/
         (height_of_node l r ((height s) + 1))) /\
       (* ...but is unchanged when no rebalancing *)
        (`-2 <= (height l) - (height r) <= 2` -> 
         (height_of_node l r (height s)))) /\
       (* elements are those of (l,x,r) *)
       (y:elt)(In_tree y s) <-> 
              ((X.eq y x) \/ (In_tree y l) \/ (In_tree y r)) }.
  Proof.
    Intros l x r bst_l avl_l bst_r avl_r; Simpl.
    Intros Hl Hr Hh.
    LetTac hl := (height l).
    LetTac hr := (height r).
    Case (Z_gt_le_dec hl (hr + 2)); Intro.
    (* hl > hr + 2 *)
    NewDestruct l.
    (* l = Leaf => absurd *)
    Simpl in hl; Unfold hl.
    Absurd hl>hr+2; Trivial.
    Generalize (height_non_negative avl_r).
    Unfold hl hr; Omega.
    (* l = Node t0 t1 t2 z0 *)
    Case (Z_ge_lt_dec (height t0) (height t2)); Intro.
    (* height t0 >= height t2 *)
    Case (create t2 x r); Auto.
    Inversion_clear bst_l; Trivial. 
    Inversion_clear avl_l; Trivial.
    Generalize Hh z; Clear Hh z; Simpl in hl; Unfold hl hr.
    AVL avl_l; AVL avl_r; Intuition Try Omega.
    Intro t2xr; Intuition.
    Case (create t0 t1 t2xr).
    Inversion_clear bst_l; Trivial. 
    Inversion_clear avl_l; Trivial.
    Intuition.
    Intuition.
    Inversion_clear bst_l; Trivial.     
    Inversion_clear bst_l; Trivial. 
    Clear H2; Intro; Intro; Intuition; Generalize (H5 y); Intuition.
    Apply ME.lt_eq with x; Auto. 
    Apply E.lt_trans with x; Auto.
    Apply Hl; Auto.
    Apply Hr; Auto.
    Clear H5.
    Generalize z H H0; Clear z H H0; Simpl in hl; Unfold hl hr.
    Unfold height_of_node in H2; AVL avl_l; AVL H3; Omega.
    Intros s (s_bst,(s_avl,(Hs1,Hs2))).
    Exists s; Simpl.
    Do 3 (Split; Trivial).
    Unfold height_of_node; Simpl.
    Clear H5 Hs2.
    Generalize z H H0; Clear z H H0; Simpl in hl; Unfold hl hr.
    Unfold height_of_node in H2 Hs1; AVL avl_l; AVL H3; AVL s_avl; Omega.
    Intuition; Generalize (Hs2 y); Generalize (H5 y); Clear Hs2 H5; Intuition.
    Inversion_clear H4; Intuition.
    (* height t0 < height t2 *)
    NewDestruct t2.
    (* t2 = Leaf => absurd *)
    Simpl in z1.
    Absurd (height t0)<0; Trivial.
    Inversion_clear avl_l; AVL H; Omega.
    (* t2 = Node t2 t3 t4 z2 *)
    Case (create t4 x r); Auto.
    Inversion_clear bst_l; Inversion_clear H0; Auto.
    Inversion_clear avl_l; Inversion_clear H0; Auto.
    Generalize z Hh; Clear z Hh; Simpl in hl; Unfold hl hr.
    Simpl in z1; AVL avl_l; Simpl in H.
    Inversion_clear avl_l; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H2; Omega.
    Intros r' (r'_bst, (r'_avl, (r'_h1, r'_h2))).
    Case (create t0 t1 t2).
    Inversion_clear bst_l; Trivial.
    Inversion_clear avl_l; Trivial.
    Inversion_clear bst_l; Inversion_clear H0; Trivial.
    Inversion_clear avl_l; Inversion_clear H0; Trivial.
    Inversion_clear bst_l; Trivial.
    Inversion_clear bst_l; Intro; Intro; Apply H2; EAuto.
    Generalize z Hh; Clear z Hh; Simpl in hl; Unfold hl hr.
    Simpl in z1; AVL avl_l; Simpl in H.
    Inversion_clear avl_l; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H2; Omega.
    Intros l' (l'_bst, (l'_avl, (l'_h1, l'_h2))).
    Case (create l' t3 r'); Auto.
    Inversion_clear bst_l; Inversion_clear H0.
    Intro; Intro; Generalize (l'_h2 y); Clear l'_h2; Intuition.
    Apply ME.eq_lt with t1; Auto.
    Apply E.lt_trans with t1; [ Apply H1 | Apply H2 ]; Auto.
    Inversion_clear bst_l; Inversion_clear H0.
    Intro; Intro; Generalize (r'_h2 y); Clear r'_h2; Intuition.
    Apply ME.lt_eq with x; Auto.
    Apply E.lt_trans with x; [Apply Hl|Apply Hr]; Auto.
    Generalize z Hh; Clear z Hh; Simpl in hl; Unfold hl hr.
    Simpl in z1; AVL avl_l; Simpl in H.
    Inversion_clear avl_l; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H2; Unfold height_of_node in r'_h1 l'_h1; Omega.
    Intros s (s_bst,(s_avl,(s_h1,s_h2))).
    Exists s; Simpl; Do 3 (Split; Trivial).
    Clear r'_h2 l'_h2 s_h2.
    Generalize z Hh; Clear z Hh; Simpl in hl; Unfold hl hr.
    AVL avl_l; Inversion_clear avl_l.
    AVL H2; Unfold height_of_node in H4; Simpl in H4.
    Unfold height_of_node; Simpl.
    Unfold height_of_node in s_h1 r'_h1 l'_h1; Simpl.
    Simpl in z1; AVL r'_avl; AVL l'_avl; Simpl in H.
    Clear bst_l bst_r avl_r Hl Hr hl hr r'_bst r'_avl
      l'_bst l'_avl s_bst s_avl H1 H2; Intuition Omega. (* 9 seconds *)
    Intro y; Generalize (r'_h2 y); 
      Generalize (l'_h2 y); Generalize (s_h2 y); 
      Clear r'_h2 l'_h2 s_h2; Intuition.
    Inversion_clear H10; Intuition.
    Inversion_clear H14; Intuition.
    (* hl <= hr + 2 *)
    Case (Z_gt_le_dec hr (hl + 2)); Intro.
    (* hr > hl + 2 *)
    NewDestruct r.
    (* r = Leaf => absurd *)
    Simpl in hr; Unfold hr.
    Absurd hr>hl+2; Trivial.
    AVL avl_l; Unfold hl hr; Omega.
    (* r = Node t0 t1 t2 z0 *)
    Case (Z_ge_lt_dec (height t2) (height t0)); Intro.
    (* height t2 >= height t0 *)
    Case (create l x t0); Auto.
    Inversion_clear bst_r; Trivial. 
    Inversion_clear avl_r; Trivial.
    Generalize Hh z z0; Clear Hh z z0; Simpl in hr; Unfold hl hr.
    AVL avl_l; AVL avl_r; Intuition Try Omega.
    Intro lxt0; Intuition.
    Case (create lxt0 t1 t2); Auto.
    Inversion_clear bst_r; Trivial. 
    Inversion_clear avl_r; Trivial.
    Clear H2; Intro; Intro; Intuition; Generalize (H5 y); Intuition.
    Apply ME.eq_lt with x; Auto. 
    Apply E.lt_trans with x; [Apply Hl|Apply Hr]; Auto.
    Inversion_clear bst_r; Auto. 
    Inversion_clear bst_r; Auto. 
    Clear H5.
    Generalize z z0 H H0; Clear z z0 H H0; Simpl in hr; Unfold hl hr.
    Unfold height_of_node in H2; AVL avl_r; AVL H3; Omega.
    Intros s (s_bst,(s_avl,(Hs1,Hs2))).
    Exists s; Simpl; Do 3 (Split; Trivial).
    Unfold height_of_node; Simpl.
    Clear H5 Hs2.
    Generalize z z0 H H0; Clear z z0 H H0; Simpl in hr; Unfold hl hr.
    Unfold height_of_node in H2 Hs1; AVL avl_r; AVL H3; AVL s_avl; Omega.
    Intuition; Generalize (Hs2 y); Generalize (H5 y); Clear Hs2 H5; Intuition.
    Inversion_clear H4; Intuition.
    (* height t2 < height t0 *)
    NewDestruct t0.
    (* t0 = Leaf => absurd *)
    Simpl in z2.
    Absurd (height t2)<0; Trivial.
    Inversion_clear avl_r; AVL H0; Omega.
    (* t0 = Node t0 t3 t4 z2 *)
    Case (create l x t0); Auto.
    Inversion_clear bst_r; Inversion_clear H; Auto.
    Inversion_clear avl_r; Inversion_clear H; Auto.
    Generalize z z0 Hh; Clear z z0 Hh; Simpl in hr; Unfold hl hr.
    Simpl in z2; AVL avl_r; Simpl in H.
    Inversion_clear avl_r; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H1; Omega.
    Intros l' (l'_bst, (l'_avl, (l'_h1, l'_h2))).
    Case (create t4 t1 t2).
    Inversion_clear bst_r; Inversion_clear H; Trivial.
    Inversion_clear avl_r; Inversion_clear H; Trivial.
    Inversion_clear bst_r; Trivial.
    Inversion_clear avl_r; Trivial.
    Inversion_clear bst_r; Intro; Intro; Apply H1; EAuto.
    Inversion_clear bst_r; Trivial.
    Generalize z z0 Hh; Clear z z0 Hh; Simpl in hr; Unfold hl hr.
    Simpl in z2; AVL avl_r; Simpl in H.
    Inversion_clear avl_r; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H1; Omega.
    Intros r' (r'_bst, (r'_avl, (r'_h1, r'_h2))).
    Case (create l' t3 r'); Auto.
    Inversion_clear bst_r; Inversion_clear H.
    Intro; Intro; Generalize (l'_h2 y); Clear l'_h2; Intuition.
    Apply ME.eq_lt with x; Auto.
    Apply E.lt_trans with x; [ Apply Hl | Apply Hr ]; Auto.
    Inversion_clear bst_r; Inversion_clear H.
    Intro; Intro; Generalize (r'_h2 y); Clear r'_h2; Intuition.
    Apply ME.lt_eq with t1; Auto.
    Apply E.lt_trans with t1; [Apply H1|Apply H2]; Auto.
    Generalize z z0 Hh; Clear z z0 Hh; Simpl in hr; Unfold hl hr.
    Simpl in z2; AVL avl_r; Simpl in H.
    Inversion_clear avl_r; Unfold height_of_node in H4; Simpl in H3 H4.
    AVL H1; Unfold height_of_node in r'_h1 l'_h1; Omega.
    Intros s (s_bst,(s_avl,(s_h1,s_h2))).
    Exists s; Simpl; Do 3 (Split; Trivial).
    Clear r'_h2 l'_h2 s_h2.
    Generalize z z0 Hh; Clear z z0 Hh; Simpl in hr; Unfold hl hr.
    AVL avl_r; Inversion_clear avl_r.
    AVL H1; Unfold height_of_node in H4; Simpl in H4.
    Unfold height_of_node; Simpl.
    Unfold height_of_node in s_h1 r'_h1 l'_h1; Simpl.
    Simpl in z2; AVL r'_avl; AVL l'_avl; Simpl in H.
    Clear bst_l bst_r avl_l Hl Hr hl hr r'_bst r'_avl
      l'_bst l'_avl s_bst s_avl H1 H2; Intuition Omega. (* 9 seconds *)
    Intro y; Generalize (r'_h2 y); 
      Generalize (l'_h2 y); Generalize (s_h2 y); 
      Clear r'_h2 l'_h2 s_h2; Intuition.
    Inversion_clear H10; Intuition.
    Inversion_clear H14; Intuition.
    (* hr <= hl + 2 *)
    LetTac s := (Node l x r ((max (height l) (height r)) + 1)).
    Assert (bst s). 
    Unfold s; Auto.
    Assert (avl s). 
    Unfold s; Constructor; Auto.
    Generalize z z0; Unfold hl hr; Intros; Omega.
    Unfold height_of_node; MaxCase '(height l) '(height r); Intros; Omega.
    Exists s; Unfold s height_of_node; Simpl; Do 3 (Split; Trivial).
    Generalize z z0; Unfold hl hr; MaxCase '(height l) '(height r); Intros; Omega.
    Intuition; Inversion_clear H3; Intuition.
  Defined.

  (** * Insertion *)

  Definition add_tree : 
    (x:elt)(s:tree)(bst s) -> (avl s) ->
    { s':tree | (bst s') /\ (avl s') /\ 
                0 <= (height s')-(height s) <= 1 /\
                ((y:elt)(In_tree y s') <-> ((E.eq y x)\/(In_tree y s))) }.
  Proof.
    Induction s; Simpl; Intros.
    (* s = Leaf *)
    Exists (Node Leaf x Leaf 1); Simpl; Intuition.
    Constructor; Unfold height_of_node; Simpl; 
      Intuition Try Omega.
    Inversion_clear H1; Intuition.
    (* s = Node t0 t1 t2 *)
    Case (X.compare x t1); Intro.
    (* x < t1 *)
    Clear H0; Case H; Clear H.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intro l'; Simpl; Intuition.
    Case (bal l' t1 t2); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intro; Intro; Generalize (H5 y); Clear H5; Intuition.
    Apply ME.eq_lt with x; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H1; Auto.
    Clear H5; AVL H2; AVL H3; Intuition.
    Intros s' (s'_bst,(s'_avl,(s'_h1, s'_h2))).
    Exists s'; Simpl; Do 3 (Split ; Trivial).
    Clear s'_h2 H; Unfold height_of_node in s'_h1.
    AVL H2; AVL H3; AVL s'_avl. Omega.
    Clear s'_h1; Intro.
    Generalize (s'_h2 y) (H5 y); Clear s'_h2 H5; Intuition.
    Inversion_clear H13; Intuition.
    (* x = t1 *)
    Clear H H0.
    Exists (Node t0 t1 t2 z); Simpl; Intuition.
    Apply IsRoot; Apply E.eq_trans with x; Auto.
    (* x > t1 *)
    Clear H; Case H0; Clear H0.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intros r' (r'_bst,(r'_avl,H3)); Simpl; Intuition.
    Case (bal t0 t1 r'); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intro; Intro; Generalize (H0 y); Clear H0; Intuition.
    Inversion_clear H1; Auto.
    Intro; Intro; Generalize (H0 y); Clear H0; Intuition.
    Apply ME.lt_eq with x; Auto.
    Inversion_clear H1; Auto.
    Clear H0; AVL H2; AVL r'_avl; Intuition.
    Intros s' (s'_bst,(s'_avl,(s'_h1, s'_h2))).
    Exists s'; Simpl; Do 3 (Split; Trivial).
    Clear s'_h2 H0; Unfold height_of_node in s'_h1.
    AVL H2; AVL r'_avl; AVL s'_avl; Omega.
    Clear s'_h1; Intro.
    Generalize (s'_h2 y) (H0 y); Clear s'_h2 H0; Intuition.
    Inversion_clear H11; Intuition.
  Defined.

  Definition add : (x:elt) (s:t) { s':t | (Add x s s') }.
  Proof.
    Intros x (s,s_bst,s_avl); Unfold Add In.
    Case (add_tree x s); Trivial.
    Intros s' (s'_bst,(s'_avl,Hs')).
    Exists (t_intro s' s'_bst s'_avl); Intuition.
  Defined.

  (** * Join

      Same as [bal] but does not assumme anything regarding heights
      of [l] and [r]. Code is
<<
    let rec join l v r =
      match (l, r) with
        (Empty, _) -> add v r
      | (_, Empty) -> add v l
      | (Node(ll, lv, lr, lh), Node(rl, rv, rr, rh)) ->
          if lh > rh + 2 then bal ll lv (join lr v r) else
          if rh > lh + 2 then bal (join l v rl) rv rr else
          create l v r
>>
  *)

  Definition join : 
    (l:tree)(x:elt)(r:tree)
    (bst l) -> (avl l) -> (bst r) -> (avl r) ->
    (lt_tree x l) -> (gt_tree x r) ->
    { s:tree | (bst s) /\ (avl s) /\
               ((height_of_node l r (height s)) \/
                (height_of_node l r ((height s)+1))) /\
               ((y:elt)(In_tree y s) <-> 
                       ((X.eq y x) \/ (In_tree y l) \/ (In_tree y r))) }.
  Proof.
    Induction l; Simpl.
    (* l = Leaf *)
    Intros; Case (add_tree x r); Trivial.
    Intros s' (s'_bst,(s'_avl,Hs')); Simpl; Exists s'; Intuition.
    Unfold height_of_node; Simpl. AVL H2; AVL s'_avl; Intuition Omega.
    Ground. Ground. Inversion_clear H5. Ground.
    Intros; Clear H.
    Induction r; Simpl.
    (* r = Leaf *)
    Clear H0.
    Intros; Case (add_tree x (Node t0 t1 t2 z)); Simpl; Trivial.
    Intros s' (s'_bst,(s'_avl,Hs')); Simpl; Exists s'; Intuition.
    Unfold height_of_node; Simpl. AVL s'_avl; Intuition Omega.
    Ground. Ground. Ground. Inversion_clear H.
    (* l = Node t0 t1 t2 z and r = Node r1 t3 r0 z0 *)
    Intros.
    Case (Z_gt_le_dec z (z0+2)); Intro.
    (* z > z0+2 *)
    Clear Hrecr1 Hrecr0.
    Case (H0 x (Node r1 t3 r0 z0)); Clear H0; Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intro s'; Unfold height_of_node; Simpl; Intuition.
    Case (bal t0 t1 s'); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Inversion_clear H1; Trivial.
    Clear H0; Intro; Intro; Generalize (H9 y); Clear H9; Intuition.
    Apply ME.lt_eq with x; Auto.
    Inversion_clear H1; Auto.
    Apply X.lt_trans with x; Auto.
    Clear H9; AVL H2; Intuition Omega.
    Intro s''; Unfold height_of_node; Simpl; Intuition.
    Exists s''; Do 3 (Split; Trivial).
    Clear H5 H6 H7 H8 H9 H13; AVL H2; Clear H2; Intuition Omega.
    Clear H0 H12 H10; Ground.
    Inversion_clear H0; Ground.
    (* z <= z0 + 2 *)
    Case (Z_gt_le_dec z0 (z+2)); Intro.
    (* z0 > z+2 *)
    Clear H0 Hrecr0.
    Case Hrecr1; Clear Hrecr1; Auto.
    Inversion_clear H3; Trivial.
    Inversion_clear H4; Trivial.
    Intro s'; Unfold height_of_node; Simpl; Intuition.
    Case (bal s' t3 r0); Auto.
    Inversion_clear H3; Trivial.
    Inversion_clear H4; Trivial.
    Inversion_clear H3; Trivial.
    Clear H0; Intro; Intro; Generalize (H9 y); Clear H9; Intuition.
    Apply ME.eq_lt with x; Auto.
    Apply X.lt_trans with x; Auto.
    Inversion_clear H3; Auto.
    Clear H9; AVL H4; Intuition Omega.
    Intro s''; Unfold height_of_node; Simpl; Intuition.
    Exists s''; Do 3 (Split; Trivial).
    Clear H5 H6 H7 H8 H9 H13; AVL H4; Clear H4; Intuition Omega.
    Clear H0 H12 H10; Ground.
    Inversion_clear H0; Ground.
    (* -2 <= z-z0 <= 2 *)
    Clear H0 Hrecr0 Hrecr1.
    Case (create (Node t0 t1 t2 z) x (Node r1 t3 r0 z0)); Simpl; Intuition.
    Exists x0; Intuition.
  Defined.

  (** * Extraction of minimum and maximum element *)

  Definition remove_min :
    (s:tree)(bst s) -> (avl s) -> ~s=Leaf ->
    { r : tree * elt |
        let (s',m) = r in
        (bst s') /\ (avl s') /\
        ((height s') = (height s) \/ (height s')=(height s)-1) /\
        ((y:elt)(In_tree y s') -> (E.lt m y)) /\
        ((y:elt)(In_tree y s) <-> (E.eq y m) \/ (In_tree y s')) }.
  Proof.
    Induction s; Simpl; Intros.
    Elim H1; Trivial.
    (* s = Node t0 t1 t2 *)
    NewDestruct t0.
    (* t0 = Leaf *)
    Clear H H0.
    Exists (t2, t1); Intuition.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    AVL H2; Simpl in H; Inversion_clear H2; AVL H5; Intuition; Omega.
    Inversion_clear H1; Apply H6; Auto.
    Inversion_clear H; Auto; Inversion_clear H0.
    (* t0 = Node t0 t3 t4 *)
    Clear H0; Case H; Clear H.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Discriminate.
    Intros (l',m); Intuition.
    Case (bal l' t1 t2); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intro; Intros; Generalize (H7 y) (H5 y); Clear H7 H5 H0; Intuition.
    Elim (!ME.eq_not_gt y m); Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H1; Trivial.
    Clear H5 H7.
    AVL H2; Omega. 
    Intro s'; Unfold height_of_node; Intuition.
    Exists (s',m).
    Do 3 (Split; Trivial).
    Clear H5 H7 H11; AVL H2; AVL H4; AVL H9; Omega.
    Clear H0 H10 H8; Intuition.
    Generalize (H5 y) (H7 y) (H11 y); Clear H5 H11; Intuition.
    Apply ME.lt_eq with t1; Auto.
    Generalize (H7 m); Inversion_clear H1; Intuition.
    Apply X.lt_trans with t1; Auto.
    Inversion_clear H1; Apply H18; Ground.
    Inversion_clear H1; Auto.
    Inversion_clear H0; Ground.
    Apply InLeft; Ground.
    Ground.
  Defined.

  Definition remove_max :
    (s:tree)(bst s) -> (avl s) -> ~s=Leaf ->
    { r : tree * elt |
        let (s',m) = r in
        (bst s') /\ (avl s') /\
        ((height s') = (height s) \/ (height s')=(height s)-1) /\
        ((y:elt)(In_tree y s') -> (E.lt y m)) /\
        ((y:elt)(In_tree y s) <-> (E.eq y m) \/ (In_tree y s')) }.
  Proof.
    Induction s; Simpl; Intros.
    Elim H1; Trivial.
    (* s = Node t0 t1 t2 *)
    NewDestruct t2.
    (* t2 = Leaf *)
    Clear H H0.
    Exists (t0, t1); Intuition.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    AVL H2; Simpl in H; Inversion_clear H2; AVL H4; Intuition; Omega.
    Inversion_clear H1; Apply H5; Auto.
    Inversion_clear H; Auto; Inversion_clear H0.
    (* t2 = Node t2 t3 t4 *)
    Clear H; Case H0; Clear H0.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Discriminate.
    Intros (r',m); Intuition.
    Case (bal t0 t1 r'); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Inversion_clear H1; Auto.
    Intro; Intros; Generalize (H7 y) (H5 y); Clear H7 H5 H0; Intuition.
    Elim (!ME.eq_not_lt y m); Auto.
    Inversion_clear H1; Auto.
    Clear H5 H7.
    AVL H2; Omega. 
    Intro s'; Unfold height_of_node; Intuition.
    Exists (s',m).
    Do 3 (Split; Trivial).
    Clear H5 H7 H11; AVL H2; AVL H4; AVL H9; Omega.
    Clear H0 H10 H8; Intuition.
    Generalize (H5 y) (H7 y) (H11 y); Clear H5 H11; Intuition.
    Apply ME.eq_lt with t1; Auto.
    Generalize (H7 m); Inversion_clear H1; Intuition.
    Apply X.lt_trans with t1; Auto.
    Inversion_clear H1; Apply H18; Ground.
    Inversion_clear H1; Ground.
    Inversion_clear H0; Ground.
    Apply InRight; Ground.
    Ground.
  Defined.

  (** * Merging two trees

    [merge t1 t2] builds the union of [t1] and [t2] assuming all elements
    of [t1] to be smaller than all elements of [t2], and
    [|height t1 - height t2| <= 2]. Code is
<<
    let merge t1 t2 =
      match (t1, t2) with
        (Empty, t) -> t
      | (t, Empty) -> t
      | (_, _) -> let (m,t'2) = remove_min t2 in bal t1 m t'2
>> 
  *)

  Definition merge :
    (s1,s2:tree)(bst s1) -> (avl s1) -> (bst s2) -> (avl s2) ->
    ((y1,y2:elt)(In_tree y1 s1) -> (In_tree y2 s2) -> (X.lt y1 y2)) ->
    `-2 <= (height s1) - (height s2) <= 2` ->
    { s:tree | (bst s) /\ (avl s) /\
               ((height_of_node s1 s2 (height s)) \/
                (height_of_node s1 s2 ((height s)+1))) /\
               ((y:elt)(In_tree y s) <-> 
                       ((In_tree y s1) \/ (In_tree y s2))) }.
  Proof.
    Destruct s1; Simpl.
    (* s1 = Leaf *)
    Intros; Exists s2; Unfold height_of_node; Simpl; Intuition.
    AVL H2; Omega.
    Inversion_clear H7.
    (* s1 = Node t0 t1 t2 *)
    Destruct s2; Simpl.
    (* s2 = Leaf *)
    Intros; Exists (Node t0 t1 t2 z); Unfold height_of_node; Simpl; Intuition.
    AVL H0; Omega.
    Inversion_clear H7.
    (* s2 = Node t3 t4 t5 *)
    Intros.
    Case (remove_min (Node t3 t4 t5 z0)); Auto.
    Discriminate.
    Intros (s'2,m); Intuition.
    Case (bal (Node t0 t1 t2 z) m s'2); Auto.
    Ground.
    Clear H3 H11; AVL H0; AVL H2; AVL H8; Simpl in H7; Omega.
    Intro s'; Unfold height_of_node; Intuition. 
    Exists s'.
    Do 3 (Split; Trivial).
    Clear H3 H9 H11 H15; AVL H0; AVL H2; AVL H8; AVL H13.
    Simpl in H7 H14 H12; Simpl; Intuition Omega.
    Clear H7 H12 H14; Ground.
  Defined.

  (** Variant where we remove from the biggest subtree *)

  Definition merge_bis :
    (s1,s2:tree)(bst s1) -> (avl s1) -> (bst s2) -> (avl s2) ->
    ((y1,y2:elt)(In_tree y1 s1) -> (In_tree y2 s2) -> (X.lt y1 y2)) ->
    `-2 <= (height s1) - (height s2) <= 2` ->
    { s:tree | (bst s) /\ (avl s) /\
               ((height_of_node s1 s2 (height s)) \/
                (height_of_node s1 s2 ((height s)+1))) /\
               ((y:elt)(In_tree y s) <-> 
                       ((In_tree y s1) \/ (In_tree y s2))) }.
  Proof.
    Destruct s1; Simpl.
    (* s1 = Leaf *)
    Intros; Exists s2; Unfold height_of_node; Simpl; Intuition.
    AVL H2; Omega.
    Inversion_clear H7.
    (* s1 = Node t0 t1 t2 *)
    Destruct s2; Simpl.
    (* s2 = Leaf *)
    Intros; Exists (Node t0 t1 t2 z); Unfold height_of_node; Simpl; Intuition.
    AVL H0; Omega.
    Inversion_clear H7.
    (* s2 = Node t3 t4 t5 *)
    Intros; Case (Z_ge_lt_dec z z0); Intro.
    (* z >= z0 *)
    Case (remove_max (Node t0 t1 t2 z)); Auto.
    Discriminate.
    Intros (s'1,m); Intuition.
    Case (create s'1 m (Node t3 t4 t5 z0)); Auto.
    Ground.
    Clear H3 H11; AVL H0; AVL H2; AVL H8; Simpl in H7; Omega.
    Intro s'; Unfold height_of_node; Intuition. 
    Exists s'.
    Do 3 (Split; Trivial).
    Clear H3 H9 H11 H15; AVL H0; AVL H2; AVL H8; AVL H13.
    Simpl in H7 H14 H12; Simpl.
    Intuition Omega.
    Clear H7 H12; Ground.
    (* z < z0 *)
    Case (remove_min (Node t3 t4 t5 z0)); Auto.
    Discriminate.
    Intros (s'2,m); Intuition.
    Case (create (Node t0 t1 t2 z) m s'2); Auto.
    Ground.
    Clear H3 H11; AVL H0; AVL H2; AVL H8; Simpl in H7; Omega.
    Intro s'; Unfold height_of_node; Intuition. 
    Exists s'.
    Do 3 (Split; Trivial).
    Clear H3 H9 H11 H15; AVL H0; AVL H2; AVL H8; AVL H13.
    Simpl in H7 H14 H12; Simpl.
    Intuition Omega.
    Clear H7 H12; Ground.
  Defined.

  (** * Deletion *)

  Definition remove_tree :
    (x:elt)(s:tree)(bst s) -> (avl s) ->
    { s':tree | (bst s') /\ (avl s') /\
                ((height s') = (height s) \/ (height s') = (height s) - 1) /\
                ((y:elt)(In_tree y s') <-> (~(E.eq y x) /\ (In_tree y s))) }.
  Proof.
    Induction s; Simpl; Intuition.
    (* s = Leaf *)
    Exists Leaf; Simpl; Intuition; 
      (Inversion_clear H1 Orelse Inversion_clear H3).
    (* s = Node t0 t1 t2 *)
    Case (X.compare x t1); Intro.
    (* x < t1 *)
    Clear H0; Case H; Clear H.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intros t'0; Intuition.
    Case (bal t'0 t1 t2); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Clear H0; Intro; Intro; Generalize (H5 y); Clear H5; Intuition.
    Inversion_clear H1; Auto.
    Inversion_clear H1; Auto.
    Clear H5; AVL H2; AVL H3; Omega.
    Intros s'; Unfold height_of_node; Intuition.
    Exists s'; Do 3 (Split; Trivial).
    Clear H5 H9; AVL H2; AVL H3; AVL H7; Omega.
    Clear H0 H8 H6; Intuition.
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
    Apply (!ME.eq_not_lt y t1); Auto.
    Apply ME.eq_lt with x; Auto.
    Apply (!ME.lt_not_gt t1 y); Auto.
    Inversion_clear H1; Apply H16; Auto.
    Apply ME.eq_lt with x; Auto.
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
    Inversion_clear H8; 
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
    (* x = t1 *)
    Clear H H0; Case (merge t0 t2).
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros; Apply X.lt_trans with t1; Inversion_clear H1; Auto.
    AVL H2; Omega.
    Intro s'; Unfold height_of_node; Intuition.
    Exists s'; Do 3 (Split; Trivial).
    Clear H5; AVL H2; AVL H3; Omega.
    Clear H0; Intro; Generalize (H5 y); Clear H5; Intuition.
    Apply (!E.lt_not_eq y t1); Auto.
    Inversion_clear H1; Apply H10; Auto.
    Apply X.eq_trans with x; Auto.
    Apply (!E.lt_not_eq t1 y); Auto.
    Inversion_clear H1; Apply H11; Auto.
    Apply X.eq_trans with x; Auto.
    Inversion_clear H8; Intuition.
    Absurd (X.eq y x); Auto.
    Apply X.eq_trans with t1; Auto.
    (* t1 < x *)
    Clear H; Case H0; Clear H0.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intros t'2; Intuition.
    Case (bal t0 t1 t'2); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Inversion_clear H1; Auto.
    Inversion_clear H1; Auto.
    Clear H0; Intro; Intro; Generalize (H5 y); Clear H5; Intuition.
    Clear H5; AVL H2; AVL H3; Omega.
    Intros s'; Unfold height_of_node; Intuition.
    Exists s'; Do 3 (Split; Trivial).
    Clear H5 H9; AVL H2; AVL H3; AVL H7; Omega.
    Clear H0 H8 H6; Intuition.
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
    Apply (!ME.eq_not_lt t1 y); Auto.
    Apply ME.lt_eq with x; Auto.
    Apply (!ME.lt_not_gt y t1); Auto.
    Inversion_clear H1; Apply H15; Auto.
    Apply ME.lt_eq with x; Auto.
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
    Inversion_clear H8; 
    Generalize (H9 y) (H5 y); Clear H5 H9; Intuition.
  Defined.

  Definition remove : (x:elt)(s:t)
                      { s':t | (y:elt)(In y s') <-> (~(E.eq y x) /\ (In y s))}.
  Proof.
    Intros x (s,Hs,Hrb); Case (remove_tree x s); Trivial.
    Intros s'; Intuition; Clear H0.
    Exists (t_intro s' H H1); Intuition.
  Defined.

 (** * Minimum element *)

  Definition min_elt : 
     (s:t){ x:elt | (In x s) /\ (For_all [y]~(E.lt y x) s) } + { Empty s }.
  Proof.
    Intros (s,Hs,Ha).
    Unfold For_all Empty In; Simpl.
    Generalize Hs; Clear Hs Ha; Induction s; Simpl; Intros.
    (* s = Leaf *)
    Right; Intros; Intro; Inversion H.
    (* s = Node c s1 t0 s0 *)
    Clear Hrecs0; Generalize Hs Hrecs1; Clear Hs Hrecs1; Case s1; Intros.
    (* s1 = Leaf *)
    Left; Exists t0; Intuition.
    Inversion_clear H.
    Absurd (X.eq x t0); Auto.
    Inversion H1.
    Inversion_clear Hs; Absurd (E.lt x t0); Auto.
    (* s1 = Node c0 t1 t2 t3 *)
    Case Hrecs1; Clear Hrecs1.
    Inversion Hs; Auto.
    (* a minimum for [s1] *)
    Intros (m,Hm); Left; Exists m; Intuition.
    Apply (H0 x); Auto.
    Assert (X.lt m t0).
    Inversion_clear Hs; Auto.
    Inversion_clear H1; Auto.
    Elim (!X.lt_not_eq x t0); [ EAuto | Auto ].
    Inversion_clear Hs.
    Elim (!ME.lt_not_gt x t0); [ EAuto | Auto ].
    (* non minimum for [s1] => absurd *)
    Intro; Right; Intuition.
    Apply (n t2); Auto.
  Defined.

  (** * Maximum element *)

  Definition max_elt : 
     (s:t){ x:elt | (In x s) /\ (For_all [y]~(E.lt x y) s) } + { Empty s }.
   Proof.
    Intros (s,Hs,Ha).
    Unfold For_all Empty In; Simpl.
    Generalize Hs; Clear Hs Ha; Induction s; Simpl; Intros.
    (* s = Leaf *)
    Right; Intros; Intro; Inversion H.
    (* s = Node c s1 t0 s0 *)
    Clear Hrecs1; Generalize Hs Hrecs0; Clear Hs Hrecs0; Case s0; Intros.
    (* s0 = Leaf *)
    Left; Exists t0; Intuition.
    Inversion_clear H.
    Absurd (X.eq x t0); Auto.
    Inversion_clear Hs; Absurd (E.lt t0 x); Auto.
    Inversion H1.
    (* s0 = Node c0 t1 t2 t3 *)
    Case Hrecs0; Clear Hrecs0.
    Inversion Hs; Auto.
    (* a maximum for [s0] *)
    Intros (m,Hm); Left; Exists m; Intuition.
    Apply (H0 x); Auto.
    Assert (X.lt t0 m).
    Inversion_clear Hs; Auto.
    Inversion_clear H1; Auto.
    Elim (!X.lt_not_eq x t0); [ EAuto | Auto ].
    Inversion_clear Hs.
    Elim (!ME.lt_not_gt t0 x); [ EAuto | Auto ].
    (* non maximum for [s0] => absurd *)
    Intro; Right; Intuition.
    Apply (n t2); Auto.
  Defined.

  (** * Any element *)

  Definition choose : (s:t) { x:elt | (In x s)} + { Empty s }.
  Proof.
    Intros (s,Hs,Ha); Unfold Empty In; Simpl; Case s; Intuition.
    Right; Intros; Inversion H.
    Left; Exists t1; Auto.
  Defined.

  (** * Concatenation

      Same as [merge] but does not assume anything about heights. Code is 
<<
    let concat t1 t2 =
      match (t1, t2) with
        (Empty, t) -> t
      | (t, Empty) -> t
      | (_, _) -> join t1 (min_elt t2) (remove_min_elt t2)
>> 
  *)

  Definition concat :
    (s1,s2:tree)(bst s1) -> (avl s1) -> (bst s2) -> (avl s2) ->
    ((y1,y2:elt)(In_tree y1 s1) -> (In_tree y2 s2) -> (X.lt y1 y2)) ->
    { s:tree | (bst s) /\ (avl s) /\
               ((y:elt)(In_tree y s) <-> 
                       ((In_tree y s1) \/ (In_tree y s2))) }.
  Proof.
    Destruct s1; Simpl.
    (* s1 = Leaf *)
    Intros; Exists s2; Simpl; Intuition.
    Inversion_clear H5.
    (* s1 = Node t0 t1 t2 *)
    Destruct s2; Simpl.
    (* s2 = Leaf *)
    Intros; Exists (Node t0 t1 t2 z); Simpl; Intuition.
    Inversion_clear H5.
    (* s2 = Node t3 t4 t5 *)
    Intros.
    Case (remove_min (Node t3 t4 t5 z0)); Auto.
    Discriminate.
    Intros (s'2,m); Intuition.
    Case (join (Node t0 t1 t2 z) m s'2); Auto.
    Ground.
    Intro s'; Intuition. 
    Exists s'.
    Do 2 (Split; Trivial).
    Clear H5 H10; Ground.
  Defined.

  (** * Splitting 

      [split x s] returns a triple [(l, present, r)] where
      - [l] is the set of elements of [s] that are [< x]
      - [r] is the set of elements of [s] that are [> x]
      - [present] is [true] if and only if [s] contains  [x].

<<
    let rec split x = function
        Empty ->
          (Empty, false, Empty)
      | Node(l, v, r, _) ->
          let c = Ord.compare x v in
          if c = 0 then (l, true, r)
          else if c < 0 then
            let (ll, pres, rl) = split x l in (ll, pres, join rl v r)
          else
            let (lr, pres, rr) = split x r in (join l v lr, pres, rr)
>> 
  *)

  Definition split :
    (x:elt)(s:tree)(bst s) -> (avl s) ->
    { res:tree*bool*tree |
      let (l,res') = res in
      let (b,r) = res' in
      (bst l) /\ (avl l) /\ (bst r) /\ (avl r) /\
      ((y:elt)(In_tree y l) <-> ((In_tree y s) /\ (X.lt y x))) /\
      ((y:elt)(In_tree y r) <-> ((In_tree y s) /\ (X.lt x y))) /\
      (b=true <-> (In_tree x s)) }.
  Proof.
    Induction s; Simpl; Intuition.
    (* s = Leaf *)
    Exists (Leaf,(false,Leaf)); Simpl; Intuition; Inversion_clear H1.
    (* s = Node t0 t1 t2 z *)
    Case (X.compare x t1); Intro.
    (* x < t1 *)
    Clear H0; Case H; Clear H.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intros (ll,(pres,rl)); Intuition.
    Case (join rl t1 t2); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Inversion_clear H1; Ground.
    Inversion_clear H1; Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (ll,(pres,s')); Intuition.
    Ground.
    Ground.
    Inversion_clear H10; Ground.
    Absurd (X.eq y t1); Auto.
    Apply E.lt_not_eq; Apply X.lt_trans with x; Auto.
    Inversion_clear H1; Absurd (X.lt y t1); Auto.
    Apply X.lt_trans with x; Auto.
    Ground.
    Generalize (s'_y y); Clear s'_y; Ground.
    Apply ME.lt_eq with t1; Auto.
    Apply X.lt_trans with t1; Inversion_clear H1; Ground.
    Inversion_clear H10; Ground.
    Inversion_clear H8; Intuition.
    Absurd (X.eq x t1); Auto.
    Absurd (X.lt x t1); Inversion_clear H1; Auto.
    (* x = t1 *)
    Clear H H0.
    Exists (t0,(true,t2)); Simpl.
    Split. Inversion_clear H1; Trivial.
    Split. Inversion_clear H2; Trivial.
    Split. Inversion_clear H1; Trivial.
    Split. Inversion_clear H2; Trivial.
    Inversion_clear H1; Ground.
    Apply ME.lt_eq with t1; Auto.
    Inversion_clear H1; Auto.
    Absurd (X.eq y t1); Auto.
    Apply E.lt_not_eq; Apply ME.lt_eq with x; Auto.
    Absurd (X.lt y x); Auto.
    Apply ME.lt_not_gt; Apply ME.eq_lt with t1; Auto.
    Apply ME.eq_lt with t1; Auto.
    Inversion_clear H1; Auto.
    Absurd (X.eq x t1); Auto.
    Apply E.lt_not_eq; Apply ME.lt_eq with y; Auto.
    Absurd (X.lt x y); Auto.
    Apply ME.lt_not_gt; Apply ME.lt_eq with t1; Auto.
    (* x > t1 *)
    Clear H; Case H0; Clear H0.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Intros (lr,(pres,rr)); Intuition.
    Case (join t0 t1 lr); Auto.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Trivial.
    Inversion_clear H1; Ground.
    Inversion_clear H1; Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (s',(pres,rr)); Inversion_clear H1; Intuition.
    Ground.
    Generalize (s'_y y); Clear s'_y; Ground.
    Apply ME.eq_lt with t1; Auto.
    Apply X.lt_trans with t1; Auto.
    Inversion_clear H13; Ground.
    Ground.
    Ground.
    Inversion_clear H13; Intuition.
    Absurd (X.lt t1 x); Auto.
    Apply ME.lt_not_gt; Apply ME.lt_eq with y; Auto.
    Absurd (X.lt x y); Auto.
    Apply ME.lt_not_gt; Apply X.lt_trans with t1; Auto.
    Ground.
    Inversion_clear H1; Auto.
    Absurd (X.eq x t1); Auto.
    Absurd (X.lt x t1); Auto.
  Defined.

  (** * Intersection
<<
    let rec inter s1 s2 =
      match (s1, s2) with
        (Empty, t2) -> Empty
      | (t1, Empty) -> Empty
      | (Node(l1, v1, r1, _), t2) ->
          match split v1 t2 with
            (l2, false, r2) ->
              concat (inter l1 l2) (inter r1 r2)
          | (l2, true, r2) ->
              join (inter l1 l2) v1 (inter r1 r2)
>> 
  *)

  Definition inter :
    (s1,s2:t)
    { s':t | (x:elt)(In x s') <-> ((In x s1)/\(In x s2))}.
  Proof.
    Intros (s1,s1_bst,s1_avl) (s2,s2_bst,s2_avl).
    Generalize s1_bst s1_avl s2 s2_bst s2_avl; Clear s1_bst s1_avl s2_bst s2_avl.
    Unfold In; Induction s1; Simpl; Intuition.
    (* s1 = Leaf *)
    Exists (t_intro Leaf s1_bst s1_avl); Simpl; Intuition.
    Inversion_clear H.
    (* s1 = Node s1_1 t0 s1_0 *)
    NewDestruct s0; Simpl.
    (* s2 = Leaf *)
    Clear Hrecs1_1 Hrecs1_0.
    Exists (t_intro Leaf s2_bst s2_avl); Simpl; Intuition.
    Inversion_clear H.
    (* s2 = Node t1 t2 t3 *)
    Case (split t0 (Node t1 t2 t3 z0)); Auto.
    Intros (l2,(b,r2)); Simpl; Intuition.
    Assert s1_1_bst: (bst s1_1). Inversion_clear s1_bst; Trivial.
    Assert s1_1_avl: (avl s1_1). Inversion_clear s1_avl; Trivial.
    Case (Hrecs1_1 s1_1_bst s1_1_avl l2 H H1); Clear Hrecs1_1.
    Intros (inter_s1_1_l2,s1_1_l2_bst,s1_1_l2_avl); Simpl; Intuition.
    Assert s1_0_bst: (bst s1_0). Inversion_clear s1_bst; Trivial.
    Assert s1_0_avl: (avl s1_0). Inversion_clear s1_avl; Trivial.
    Case (Hrecs1_0 s1_0_bst s1_0_avl r2 H0 H2); Clear Hrecs1_0.
    Intros (inter_s1_0_r2,s1_0_r2_bst,s1_0_r2_avl); Simpl; Intuition.
    Induction b; Simpl; Intuition.
    (* b = true *)
    Case (join inter_s1_1_l2 t0 inter_s1_0_r2); Auto.
    Ground.
    Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x) (a0 x) (a x) (H3 x) (H4 x); 
      Clear s'_y a0 a H3 H4; Intuition.
    Assert (In_tree x (Node t1 t2 t3 z0)); Auto.
    Apply eq_In_tree with t0; Auto.
    Inversion_clear H18; Intuition.
    Case (X.compare x t0); Intuition.
    Absurd (X.lt t0 x); Inversion_clear s1_bst; Auto.
    Case (X.compare x t0); Intuition.
    Absurd (X.lt x t0); Inversion_clear s1_bst; Auto.
    (* b = false *)
    Case (concat inter_s1_1_l2 inter_s1_0_r2); Auto.
    Intuition.
    Apply X.lt_trans with t0; Inversion_clear s1_bst; Ground.
    Intros s' (s'_bst,(s'_avl,s'_y)).
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x) (a0 x) (a x) (H3 x) (H4 x); 
      Clear s'_y a0 a H3 H4; Intuition.
    Assert ~(X.eq x t0). Intro.
    Assert false=true. Apply H7; Apply eq_In_tree with x; Auto.
    Discriminate H19.
    Inversion_clear H17; Intuition.
    Case (X.compare x t0); Intuition.
    Absurd (X.lt t0 x); Inversion_clear s1_bst; Auto.
    Case (X.compare x t0); Intuition.
    Absurd (X.lt x t0); Inversion_clear s1_bst; Auto.
  Defined.

  (** * Difference
<<
     let rec diff s1 s2 =
      match (s1, s2) with
        (Empty, t2) -> Empty
      | (t1, Empty) -> t1
      | (Node(l1, v1, r1, _), t2) ->
          match split v1 t2 with
            (l2, false, r2) ->
              join (diff l1 l2) v1 (diff r1 r2)
          | (l2, true, r2) ->
              concat (diff l1 l2) (diff r1 r2)
>> 
  *)

  Definition diff :
    (s1,s2:t)
    { s':t | (x:elt)(In x s') <-> ((In x s1) /\ ~(In x s2))}.
  Proof.
    Intros (s1,s1_bst,s1_avl) (s2,s2_bst,s2_avl).
    Generalize s1_bst s1_avl s2 s2_bst s2_avl; Clear s1_bst s1_avl s2_bst s2_avl.
    Unfold In; Induction s1; Simpl; Intuition.
    (* s1 = Leaf *)
    Exists (t_intro Leaf s1_bst s1_avl); Simpl; Intuition.
    Inversion_clear H.
    (* s1 = Node s1_1 t0 s1_0 *)
    NewDestruct s0; Simpl.
    (* s2 = Leaf *)
    Clear Hrecs1_1 Hrecs1_0.
    Exists (t_intro (Node s1_1 t0 s1_0 z) s1_bst s1_avl); Simpl; Intuition.
    Inversion_clear H0.
    (* s2 = Node t1 t2 t3 *)
    Case (split t0 (Node t1 t2 t3 z0)); Auto.
    Intros (l2,(b,r2)); Simpl; Intuition.
    Assert s1_1_bst: (bst s1_1). Inversion_clear s1_bst; Trivial.
    Assert s1_1_avl: (avl s1_1). Inversion_clear s1_avl; Trivial.
    Case (Hrecs1_1 s1_1_bst s1_1_avl l2 H H1); Clear Hrecs1_1.
    Intros (diff_s1_1_l2,s1_1_l2_bst,s1_1_l2_avl); Simpl; Intuition.
    Assert s1_0_bst: (bst s1_0). Inversion_clear s1_bst; Trivial.
    Assert s1_0_avl: (avl s1_0). Inversion_clear s1_avl; Trivial.
    Case (Hrecs1_0 s1_0_bst s1_0_avl r2 H0 H2); Clear Hrecs1_0.
    Intros (diff_s1_0_r2,s1_0_r2_bst,s1_0_r2_avl); Simpl; Intuition.
    Induction b; Simpl; Intuition.
    (* b = true *)
    Case (concat diff_s1_1_l2 diff_s1_0_r2); Auto.
    Intuition.
    Apply X.lt_trans with t0; Inversion_clear s1_bst; Ground.
    Intros s' (s'_bst,(s'_avl,s'_y)).
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x); Clear s'_y; Intuition.
    Apply InLeft; Ground.
    Clear a0 H4; Generalize (a x) (H3 x); Clear a H3; Intuition.
    Assert (X.lt x t0); Inversion_clear s1_bst; Auto.
    Apply InRight; Ground.
    Clear a H3; Generalize (a0 x) (H4 x); Clear a0 H4; Intuition.
    Assert (X.lt t0 x); Inversion_clear s1_bst; Auto.
    Inversion_clear H11; Intuition.
    Elim H12; Apply eq_In_tree with t0; Auto.
    Apply H7; Clear a0 H4; Ground.
    Apply H10; Clear a H3; Ground.
    (* b = false *)
    Case (join diff_s1_1_l2 t0 diff_s1_0_r2); Auto.
    Intro; Inversion_clear s1_bst; Ground.
    Intro; Inversion_clear s1_bst; Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Intuition.
    Generalize (s'_y x); Clear s'_y; Intuition.
    Apply InLeft; Ground.
    Apply InRight; Ground.
    Generalize (s'_y x); Clear s'_y; Intuition.
    Assert false=true. Apply H7; Apply eq_In_tree with x; Auto.
    Discriminate H12.
    Clear a0 H4; Generalize (a x) (H3 x); Clear a H3; Intuition.
    Assert (X.lt x t0); Inversion_clear s1_bst; Auto.
    Clear a H3; Generalize (a0 x) (H4 x); Clear a0 H4; Intuition.
    Assert (X.lt t0 x); Inversion_clear s1_bst; Auto.
    Generalize (s'_y x); Clear s'_y; Intuition.
    Inversion_clear H8; Intuition.
    Clear a0 H4; Generalize (a x) (H3 x); Clear a H3; Intuition.
    Clear a H3; Generalize (a0 x) (H4 x); Clear a0 H4; Intuition.
  Defined.

  (** * Elements *)

  (** [elements_tree_aux acc t] catenates the elements of [t] in infix
      order to the list [acc] *)

  Fixpoint elements_tree_aux [acc:(list X.t); t:tree] : (list X.t) :=
    Cases t of
    | Leaf => 
        acc
    | (Node l x r _) => 
        (elements_tree_aux (x :: (elements_tree_aux acc r)) l)
    end.

  (** then [elements_tree] is an instanciation with an empty [acc] *)

  Definition elements_tree := (elements_tree_aux []).

  Lemma elements_tree_aux_acc_1 : 
    (s:tree)(acc:(list elt))
    (x:elt)(InList E.eq x acc)->(InList E.eq x (elements_tree_aux acc s)).
  Proof.
    Induction s; Simpl; Intuition.
  Save.
  Hints Resolve elements_tree_aux_acc_1.

  Lemma elements_tree_aux_1 : 
    (s:tree)(acc:(list elt))
    (x:elt)(In_tree x s)->(InList E.eq x (elements_tree_aux acc s)).
  Proof.
    Induction s; Simpl; Intuition.
    Inversion H.
    Inversion_clear H1; Ground.
  Save.

  Lemma elements_tree_1 : 
    (s:tree)
    (x:elt)(In_tree x s)->(InList E.eq x (elements_tree s)).
  Proof.
    Unfold elements_tree; Intros; Apply elements_tree_aux_1; Auto.
  Save.
  Hints Resolve elements_tree_1.

  Lemma elements_tree_aux_acc_2 : 
    (s:tree)(acc:(list elt))
    (x:elt)(InList E.eq x (elements_tree_aux acc s)) ->
      (InList E.eq x acc) \/ (In_tree x s).
  Proof.
    Induction s; Simpl; Intuition.
    Elim (H ?? H1); Intuition.
    Inversion_clear H2; Intuition.
    Elim (H0 ?? H3); Intuition.
  Save.
  Hints Resolve elements_tree_aux_acc_2.

  Lemma elements_tree_2 : 
   (s:tree)
    (x:elt)(InList E.eq x (elements_tree s)) -> (In_tree x s).
  Proof.
    Unfold elements_tree; Intros.
    Elim (elements_tree_aux_acc_2 ??? H); Auto.
    Intros; Inversion H0.
  Save.
  Hints Resolve elements_tree_2.

  Lemma elements_tree_aux_sort : 
    (s:tree)(bst s) -> (acc:(list elt))
    (sort E.lt acc) -> 
    ((x:elt)(InList E.eq x acc) -> (y:elt)(In_tree y s) -> (E.lt y x)) ->
    (sort E.lt (elements_tree_aux acc s)).
  Proof.
    Induction s; Simpl; Intuition.
    Apply H.
    Inversion H1; Auto.
    Constructor. 
    Apply H0; Auto.
    Inversion H1; Auto.
    Apply ME.Inf_In_2.
    Replace X.eq with E.eq; Replace X.lt with E.lt; Auto.
    Intros.
    Elim (elements_tree_aux_acc_2 t2 acc y); Intuition.
    Inversion_clear H1.
    Apply H9; Auto.
    Intuition.
    Inversion_clear H4.
    Apply ME.lt_eq with t1; Auto.
    Inversion_clear H1.
    Apply H8; Auto.
    Elim (elements_tree_aux_acc_2 ??? H6); Intuition.
    Apply E.lt_trans with t1.
    Inversion_clear H1; Apply H9; Auto.
    Inversion_clear H1; Apply H10; Auto.
  Save.

  Lemma elements_tree_sort : 
    (s:tree)(bst s) -> (sort E.lt (elements_tree s)).
  Proof.
    Intros; Unfold elements_tree; Apply elements_tree_aux_sort; Auto.
    Intros; Inversion H0.
  Save.
  Hints Resolve elements_tree_sort.

  Definition elements : 
    (s:t){ l:(list elt) | (sort E.lt l) /\ (x:elt)(In x s)<->(InList E.eq x l)}.
  Proof.
    Intros (s,Hs,Hrb); Unfold In; Simpl.
    Exists (elements_tree s); Repeat Split.
    Apply elements_tree_sort; Auto.
    Apply elements_tree_1; Auto.
    Apply elements_tree_2; Auto.
  Defined.

  (** * Cardinal *)

  Fixpoint cardinal_tree [s:tree] : nat := Cases s of
    | Leaf => O
    | (Node l _ r _) => (S (plus (cardinal_tree l) (cardinal_tree r))) end.

  Lemma cardinal_elements_1 : 
    (s:tree)(acc:(list X.t))
    (plus (length acc) (cardinal_tree s)) =
    (length (elements_tree_aux acc s)).
  Proof.
    Induction s; Simpl; Intuition.
    Rewrite <- H.
    Simpl.
    Rewrite <- H0; Omega.
  Save.

  Lemma cardinal_elements_2 :
    (s:tree)(cardinal_tree s)=(length (elements_tree s)).
  Proof.
    Exact [s](cardinal_elements_1 s []).
  Save.

  Definition cardinal : 
    (s:t) { r : nat | (EX l:(list elt) | 
                              (Unique E.eq l) /\ 
                              ((x:elt)(In x s)<->(InList E.eq x l)) /\ 
                              r = (length l)) }.
  Proof.
    Intros (s,s_bst,s_avl); Unfold In; Simpl; Clear s_avl.
    Exists (cardinal_tree s).
    Exists (elements_tree s); Intuition.
    Apply ME.Sort_Unique; Auto.
    Exact (cardinal_elements_2 s).
  Save.

  (** Induction over cardinals *)

  Lemma sorted_subset_cardinal :
    (l',l:(list X.t))(sort X.lt l) -> (sort X.lt l') ->
    ((x:elt)(InList X.eq x l) -> (InList X.eq x l')) ->
    (le (length l) (length l')).
  Proof.
    Induction l'; Simpl; Intuition.
    NewDestruct l; Trivial; Intros.
    Absurd (InList X.eq t0 []); Intuition.
    Inversion_clear H2.
    Inversion_clear H1.
    NewDestruct l0; Simpl; Intuition.
    Inversion_clear H0.
    Apply le_n_S.
    Case (X.compare t0 a); Intro.
    Absurd (InList X.eq t0 a::l).
    Intro.
    Inversion_clear H0.
    Apply (!X.lt_not_eq t0 a); Auto.
    Assert (X.lt a t0).
    Apply ME.In_sort with l; Auto.
    Apply (!ME.lt_not_gt t0 a); Auto.
    Ground.
    Apply H; Auto.
    Intros.
    Assert (InList X.eq x a::l).
    Apply H2; Auto.
    Inversion_clear H6; Auto.
    Assert (X.lt t0 x).
    Apply ME.In_sort with l0; Auto.
    Elim (!X.lt_not_eq t0 x); Auto.
    Apply X.eq_trans with a; Auto.
    Apply le_trans with (length t0::l0).
    Simpl; Omega.
    Apply (H t0::l0); Auto.
    Intros.
    Assert (InList X.eq x a::l); Ground.
    Inversion_clear H6; Auto.
    Assert (X.lt a x).
    Apply ME.In_sort with t0::l0; Auto.
    Elim (!X.lt_not_eq a x); Auto.
  Save.

  Lemma cardinal_subset : 
    (a,b:tree)(bst a) -> (bst b) ->
    ((y:elt)(In_tree y a) -> (In_tree y b)) ->
    (le (cardinal_tree a) (cardinal_tree b)).
  Proof.
    Intros.
    Do 2 Rewrite cardinal_elements_2.
    Apply sorted_subset_cardinal; Auto.
  Save.

  Lemma cardinal_left :
    (l,r:tree)(x:elt)(h:Z)
    (lt (cardinal_tree l) (cardinal_tree (Node l x r h))).
  Proof.
    Simpl; Intuition.
  Save. 

  Lemma cardinal_right :
    (l,r:tree)(x:elt)(h:Z)
    (lt (cardinal_tree r) (cardinal_tree (Node l x r h))).
  Proof.
    Simpl; Intuition.
  Save. 

  Lemma cardinal_rec2 : 
    (P:tree->tree->Set)
    ((x,x':tree)
       ((y,y':tree)(lt (plus (cardinal_tree y) (cardinal_tree y')) 
                       (plus (cardinal_tree x) (cardinal_tree x'))) ->(P y y')) 
       -> (P x x')) ->
    (x,x':tree)(P x x').
  Proof.
    Intros P H x x'.
    Apply well_founded_induction_type_2 
      with R:=[yy',xx':tree*tree]
              (lt (plus (cardinal_tree (Fst yy')) (cardinal_tree (Snd yy'))) 
                  (plus (cardinal_tree (Fst xx')) (cardinal_tree (Snd xx'))));
    Auto.                      
    Apply (Wf_nat.well_founded_ltof ? 
           [xx':tree*tree](plus (cardinal_tree (Fst xx')) (cardinal_tree (Snd xx')))).
(***
    Intros P H.
    Assert (n:nat)(x,x':tree)n=(plus (cardinal_tree x) (cardinal_tree x'))->(P x x').
    Intros n; Pattern n; Apply Wf_nat.lt_wf_rec.
    Intuition.
    Apply H; Intros.
    Apply (H0 (plus (cardinal_tree y) (cardinal_tree y'))).
    Omega.
    Omega.
    Intros; Apply (H0 (plus (cardinal_tree x) (cardinal_tree x'))); Trivial.
***)
  Save.

  Lemma height_0 : (s:tree)(avl s) -> (height s)=0 -> (x:elt)~(In_tree x s).
  Proof.
    Destruct 1; Intuition.
    Inversion_clear H1.
    Unfold height_of_node in H3.
    AVL H; AVL H0; AVL H1; Simpl in H4; Omega.
  Save.
  Implicits height_0.

  (** * Union

      [union s1 s2] does an induction over the sum of the cardinals of
      [s1] and [s2]. Code is
<<
    let rec union s1 s2 =
      match (s1, s2) with
        (Empty, t2) -> t2
      | (t1, Empty) -> t1
      | (Node(l1, v1, r1, h1), Node(l2, v2, r2, h2)) ->
          if h1 >= h2 then
            if h2 = 1 then add v2 s1 else begin
              let (l2, _, r2) = split v1 s2 in
              join (union l1 l2) v1 (union r1 r2)
            end
          else
            if h1 = 1 then add v1 s2 else begin
              let (l1, _, r1) = split v2 s1 in
              join (union l1 l2) v2 (union r1 r2)
            end
>>
  *)

  Definition union : (s1,s2:t)
                    { s':t | (x:elt)(In x s') <-> ((In x s1)\/(In x s2))}.
  Proof.
    Intros (s1,s1_bst,s1_avl) (s2,s2_bst,s2_avl).
    Generalize s1_bst s1_avl s2_bst s2_avl; Clear s1_bst s2_bst.
    Pattern s1 s2; Apply cardinal_rec2.
    Unfold In; Destruct x; Simpl; Intuition.
    (* x = Leaf *)
    Clear H.
    Exists (t_intro x' s2_bst s2_avl0); Simpl; Intuition.
    Inversion_clear H0.
    (* x = Node t0 t1 t2 *)
    NewDestruct x'; Simpl.
    (* x' = Leaf *)
    Clear H.
    Exists (t_intro (Node t0 t1 t2 z) s1_bst s1_avl0); Simpl; Intuition.
    Inversion_clear H0.
    (* x' = Node t3 t4 t5 *)
    Case (Z_ge_lt_dec z z0); Intro.
    (* z >= z0 *)
    Case (Z_eq_dec z0 1); Intro.
    (* z0 = 1 *)
    Clear H.
    Case (add_tree t4 (Node t0 t1 t2 z)); Simpl; Auto.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))).
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x0); Clear s'_y; Intuition.
    Inversion_clear H5; Intuition.
    Assert (height t3)=0.
    Inversion s2_avl0; AVL s2_avl0; AVL H9; AVL H10; Omega.
    Inversion_clear s2_avl0; Elim (height_0 H6 H5 H3).
    Assert (height t5)=0.
    Inversion s2_avl0; AVL s2_avl0; AVL H9; AVL H10; Omega.
    Inversion_clear s2_avl0; Elim (height_0 H7 H5 H3).
    (* z0 <> 1 *)
    Intros.
    Case (split t1 (Node t3 t4 t5 z0)); Auto.
    Intros (l2,(b,r2)); Simpl; Intuition.
    Assert t0_bst: (bst t0). Inversion_clear s1_bst; Trivial.
    Assert t0_avl: (avl t0). Inversion_clear s1_avl0; Trivial.
    Case (H t0 l2); Trivial.
    Assert (le (cardinal_tree l2) (cardinal_tree (Node t3 t4 t5 z0))).
    Apply cardinal_subset; Trivial.
    Ground. Omega.
    Intros (union_t0_l2,t0_l2_bst,t0_l2_avl); Simpl; Intuition.
    Assert t2_bst: (bst t2). Inversion_clear s1_bst; Trivial.
    Assert t2_avl: (avl t2). Inversion_clear s1_avl0; Trivial.
    Case (H t2 r2); Trivial.
    Assert (le (cardinal_tree r2) (cardinal_tree (Node t3 t4 t5 z0))).
    Apply cardinal_subset; Trivial.
    Ground. Omega.
    Intros (union_t2_r2,t2_r2_bst,t2_r2_avl); Simpl; Intuition.
    Case (join union_t0_l2 t1 union_t2_r2); Auto.
    Inversion_clear s1_bst; Ground.
    Inversion_clear s1_bst; Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x0) (a0 x0) (a x0) (H4 x0) (H5 x0); 
      Clear s'_y a0 a H4 H5; Intuition.
    Inversion_clear H21; Intuition.
    Case (X.compare x0 t1); Intuition.
    (* z < z0 *)
    Case (Z_eq_dec z 1); Intro.
    (* z = 1 *)
    Case (add_tree t1 (Node t3 t4 t5 z0)); Simpl; Auto.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))).
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x0); Clear s'_y; Intuition.
    Inversion_clear H6; Intuition.
    Assert (height t0)=0.
    Inversion s1_avl0; AVL s1_avl0; AVL H10; AVL H11; Omega.
    Inversion_clear s1_avl0; Elim (height_0 H7 H6 H4).
    Assert (height t2)=0.
    Inversion s1_avl0; AVL s1_avl0; AVL H10; AVL H11; Omega.
    Inversion_clear s1_avl0; Elim (height_0 H8 H6 H4).
    (* z <> 1 *)
    Intros.
    Case (split t4 (Node t0 t1 t2 z)); Auto.
    Intros (l1,(b,r1)); Simpl; Intuition.
    Assert t3_bst: (bst t3). Inversion_clear s2_bst; Trivial.
    Assert t3_avl: (avl t3). Inversion_clear s2_avl0; Trivial.
    Case (H l1 t3); Trivial.
    Assert (le (cardinal_tree l1) (cardinal_tree (Node t0 t1 t2 z))).
    Apply cardinal_subset; Trivial.
    Ground. Simpl in H7; Simpl; Omega.
    Intros (union_l1_t3,l1_t3_bst,l1_t3_avl); Simpl; Intuition.
    Assert t5_bst: (bst t5). Inversion_clear s2_bst; Trivial.
    Assert t5_avl: (avl t5). Inversion_clear s2_avl0; Trivial.
    Case (H r1 t5); Trivial.
    Assert (le (cardinal_tree r1) (cardinal_tree (Node t0 t1 t2 z))).
    Apply cardinal_subset; Trivial.
    Ground. Simpl in H7; Simpl; Omega.
    Intros (union_r1_t5,r1_t5_bst,r1_t5_avl); Simpl; Intuition.
    Case (join union_l1_t3 t4 union_r1_t5); Auto.
    Inversion_clear s2_bst; Ground.
    Inversion_clear s2_bst; Ground.
    Intros s' (s'_bst,(s'_avl,(s'_h,s'_y))); Clear s'_h.
    Exists (t_intro s' s'_bst s'_avl); Simpl; Intros.
    Generalize (s'_y x0) (a0 x0) (a x0) (H4 x0) (H5 x0); 
      Clear s'_y a0 a H4 H5; Intuition.
    Case (X.compare x0 t4); Intuition.
    Inversion_clear H21; Intuition.
  Defined.

  (** * Filter
<<
    let filter p s =
      let rec filt accu = function
        | Empty -> accu
        | Node(l, v, r, _) ->
            filt (filt (if p v then add v accu else accu) l) r in
      filt Empty s
>> 
  *)

  Definition filter_acc :
     (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})
     (s,acc:tree)(bst s) -> (avl s) -> (bst acc) -> (avl acc) ->
     { s':tree | 
         (bst s') /\ (avl s') /\
         ((compat_P E.eq P) -> 
          (x:elt)(In_tree x s') <-> 
                 ((In_tree x acc) \/ ((In_tree x s)/\(P x)))) }.
  Proof.
    Induction s; Simpl; Intuition.
    (* s = Leaf *)
    Exists acc; Intuition.
    Inversion_clear H4.
    (* s = Node t0 t1 t2 *)
    Case (Pdec t1); Intro.
    (* P t1 *)
    Case (add_tree t1 acc); Auto.
    Intros acc'; Intuition; Clear H8 H10.
    Case (H acc'); Clear H; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros acc''; Intuition.
    Case (H0 acc''); Clear H0; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros s'; Intuition; Exists s'; Do 2 (Split; Trivial).
    Intros HP x; Generalize (H9 x) (H10 HP x) (H12 HP x);
      Clear H9 H10 H12; Intuition.
    Right; Split.
    Apply IsRoot; Auto.
    Unfold compat_P in HP; Apply HP with t1; Auto.
    Inversion_clear H18; Intuition.
    (* ~(P t1) *)
    Case (H acc); Clear H; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros acc'; Intuition.
    Case (H0 acc'); Clear H0; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros s'; Intuition; Exists s'; Do 2 (Split; Trivial).
    Intros HP x; Generalize (H7 HP x) (H9 HP x);
      Clear H7 H9; Intuition.
    Inversion_clear H13; Intuition.
    Absurd (P t1); Auto.
    Unfold compat_P in HP; Apply HP with x; Auto.
  Save.

  Definition filter : (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})(s:t)
     { s':t | (compat_P E.eq P) -> (x:elt)(In x s') <-> ((In x s)/\(P x)) }.
  Proof.
    Intros P Pdec (s,s_bst,s_avl).
    Case (filter_acc P Pdec s Leaf); Auto.
    Intros s'; Intuition.
    Exists (t_intro s' H H1); Unfold In; Simpl; Intros.
    Generalize (H2 H0 x); Clear H2; Intuition;
    Inversion_clear H3.
  Save.

  (** * Partition
<<
    let partition p s =
      let rec part (t, f as accu) = function
        | Empty -> accu
        | Node(l, v, r, _) ->
            part (part (if p v then (add v t, f) else (t, add v f)) l) r in
      part (Empty, Empty) s
>>
  *)

  Definition partition_acc :
    (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})
    (s,acct,accf:tree)
    (bst s) -> (avl s) -> 
    (bst acct) -> (avl acct) -> (bst accf) -> (avl accf) ->
    { partition : tree*tree | 
        let (s1,s2) = partition in 
        (bst s1) /\ (avl s1) /\ (bst s2) /\ (avl s2) /\
	((compat_P E.eq P) -> 
	 (x:elt)(((In_tree x s1) <-> 
                 ((In_tree x acct) \/ ((In_tree x s) /\ (P x)))) /\ 
	        (((In_tree x s2) <-> 
                 ((In_tree x accf) \/ ((In_tree x s) /\ ~(P x))))))) }.
  Proof.
    Induction s; Simpl; Intuition.
    (* s = Leaf *)
    Exists (acct,accf); Simpl; Intuition;
    Inversion_clear H6.
    (* s = Node t0 t1 t2 *)
    Case (Pdec t1); Intro.
    (* P t1 *)
    Case (add_tree t1 acct); Auto.
    Intro acct'; Intuition; Clear H10 H12.
    Case (H acct' accf); Clear H; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros (acct'',accf'); Intuition.
    Case (H0 acct'' accf'); Clear H0; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros (s1,s2); Intuition; Exists (s1,s2); Do 4 (Split; Trivial).
    Intros HP x; Generalize (H11 x) (H14 HP x) (H18 HP x);
      Clear H11 H14 H18; Intros.
    Decompose [and] H11; Clear H11.
    Decompose [and] H14; Clear H14.
    Decompose [and] H17; Clear H17.
    Repeat Split; Intros.
    Elim (H24 H17); Intros.
    Elim (H21 H20); Intros.
    Elim (H18 H27); Intros.
    Right; Split; Auto.
    Apply (HP t1); Auto.
    Auto.
    Right; Decompose [and] H27; Auto.
    Right; Decompose [and] H20; Auto.
    Elim H17; Intros.
    Apply H25; Left; Apply H22; Left; Apply H19; Right; Trivial.
    Decompose [and] H20; Clear H20.
    Inversion_clear H27.
    Apply H25; Left; Apply H22; Left; Apply H19; Left; Trivial.
    Apply H25; Left; Apply H22; Right; Auto.
    Apply H25; Right; Auto.
    Elim (H14 H17); Intros.
    Elim (H11 H20); Intros.
    Auto.
    Right; Decompose [and] H27; Auto.
    Right; Decompose [and] H20; Auto.
    Elim H17; Intros.
    GIntuition.
    GIntuition.
    Inversion_clear H17; GIntuition; Elim H33; Apply (HP t1); Auto.
    (* ~(P t1) *)
    Case (add_tree t1 accf); Auto.
    Intro accf'; Intuition; Clear H10 H12.
    Case (H acct accf'); Clear H; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros (acct',accf''); Intuition.
    Case (H0 acct' accf''); Clear H0; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Auto.
    Intros (s1,s2); Intuition; Exists (s1,s2); Do 4 (Split; Trivial).
    Intros HP x; Generalize (H11 x) (H14 HP x) (H18 HP x);
      Clear H11 H14 H18; Intros.
    Decompose [and] H11; Clear H11.
    Decompose [and] H14; Clear H14.
    Decompose [and] H17; Clear H17.
    Repeat Split; Intros.
    Elim (H24 H17); Intros.
    Elim (H21 H20); Intros.
    Auto.
    Right; Decompose [and] H27; Auto.
    Right; Decompose [and] H20; Auto.
    Elim H17; Intros.
    Apply H25; Left; Apply H22; Left; Auto.
    Decompose [and] H20; Clear H20.
    Inversion_clear H27.
    Elim f; Apply (HP x); Auto.
    GIntuition.
    GIntuition.
    Elim (H14 H17); Intros.
    Elim (H11 H20); Intros.
    Elim (H18 H27); Auto; Intros.
    Right; Split; Auto; Intro; Apply f; Apply (HP x); Auto.
    Right; Decompose [and] H27; Auto.
    Right; Decompose [and] H20; Auto.
    GIntuition.
    Inversion_clear H17; GIntuition; Elim H33; Apply (HP t1); Auto.
  Defined.

  Definition partition : 
    (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})(s:t)
    { partition : t*t | 
        let (s1,s2) = partition in 
	(compat_P E.eq P) -> 
	((For_all P s1) /\ 
	 (For_all ([x]~(P x)) s2) /\
	 (x:elt)(In x s)<->((In x s1)\/(In x s2))) }.
  Proof.
    Intros P Pdec (s,s_bst,s_avl).
    Case (partition_acc P Pdec s Leaf Leaf); Auto.
    Intros (s1,s2); Intuition.
    Exists ((t_intro s1 H H1), (t_intro s2 H0 H2)); Unfold In; Simpl; Intros.
    Unfold For_all In; Simpl.
    Split; [Idtac|Split]; Intro x; Generalize (H4 H3 x); Clear H4; Intuition.
    Inversion_clear H4.
    Inversion_clear H4.
    Inversion_clear H7.
    Inversion_clear H7.
    Case (Pdec x); Auto.
    Inversion_clear H5.
    Inversion_clear H4.
  Save.

  (** * Subset
<<
    let rec subset s1 s2 =
      match (s1, s2) with
        Empty, _ ->
          true
      | _, Empty ->
          false
      | Node (l1, v1, r1, _), (Node (l2, v2, r2, _) as t2) ->
          let c = Ord.compare v1 v2 in
          if c = 0 then
            subset l1 l2 && subset r1 r2
          else if c < 0 then
            subset (Node (l1, v1, Empty, 0)) l2 && subset r1 t2
          else
            subset (Node (Empty, v1, r1, 0)) r2 && subset l1 t2
>>
  *)

  Definition subset :
    (s1,s2:t){ Subset s1 s2 } + { ~(Subset s1 s2) }.
  Proof.
    Unfold Subset In.
    Intros (s1,s1_bst,s1_avl) (s2,s2_bst,s2_avl); Simpl.
    Generalize s1_bst s2_bst; Clear s1_bst s1_avl s2_bst s2_avl.
    Pattern s1 s2; Apply cardinal_rec2.
    Destruct x; Simpl; Intuition.
    (* x = Leaf *)
    Clear H; Left; Intros; Inversion_clear H.
    (* x = Node t0 t1 t2 z *)
    NewDestruct x'; Simpl; Intuition.
    (* x' = Leaf *)
    Right; Intros.
    Assert (In_tree t1 Leaf); Auto.
    Inversion_clear H1.
    (* x' = Node t3 t4 t5 z0 *)
    Case (X.compare t1 t4); Intro.
    (* t1 < t4 *)
    Case (H (Node t0 t1 Leaf 0) t3); Auto; Intros.
    Simpl; Omega.
    Constructor; Inversion_clear s1_bst; Auto.
    Inversion_clear s2_bst; Auto.
    (* Subset (Node t0 t1 Leaf) t3 *)
    Intros; Case (H t2 (Node t3 t4 t5 z0)); Auto; Intros.
    Simpl; Omega.
    Inversion_clear s1_bst; Auto.
    (* Subset t2 (Node t3 t4 t5 z0) *)
    Clear H; Left; Intuition.
    Generalize (i a) (i0 a); Clear i i0; Inversion_clear H; Intuition.
    (* ~ (Subset t2 (Node t3 t4 t5 z0)) *)
    Clear H; Right; Intuition.
    Apply f; Intuition.
    Assert (In_tree a (Node t3 t4 t5 z0)).
    Apply H; Inversion_clear H0; Auto.
    Inversion_clear H1.
    Inversion_clear H1; Auto.
    Inversion_clear H0; Auto.
    Elim (!X.lt_not_eq t1 t4); Auto.
    Apply X.eq_trans with a; Auto.
    Assert (X.lt a t1).
    Inversion_clear s1_bst; Apply H4; Auto.
    Elim (!X.lt_not_eq a t4); Auto.
    Apply X.lt_trans with t1; Auto.
    Inversion_clear H1.
    Assert (X.lt t4 a).
    Inversion_clear s2_bst; Apply H5; Auto.
    Inversion_clear H0; Auto.
    Elim (!X.lt_not_eq t1 a); Auto.
    Apply X.lt_trans with t4; Auto.
    Assert (X.lt a t1).
    Inversion_clear s1_bst; Apply H5; Auto.
    Elim (!ME.lt_not_gt a t1); Auto.
    Apply X.lt_trans with t4; Auto.
    Inversion_clear H3.
    (* t1 = t4 *)
    Case (H t0 t3); Auto; Intros.
    Simpl; Omega.
    Inversion_clear s1_bst; Auto.
    Inversion_clear s2_bst; Auto.
    (* Subset t0 t3 *)
    Case (H t2 t5); Auto; Intros.
    Simpl; Omega.
    Inversion_clear s1_bst; Auto.
    Inversion_clear s2_bst; Auto.
    (* Subset t2 t5 *)
    Clear H; Left; Intuition.
    Inversion_clear H; Intuition.
    Apply IsRoot; Apply X.eq_trans with t1; Auto.
    (* ~(Subset t2 t5) *)
    Clear H; Right; Intuition.
    Apply f; Intros.
    Assert (In_tree a (Node t3 t4 t5 z0)).
    Apply H; Auto.
    Inversion_clear H1; Auto.
    Elim (!X.lt_not_eq t1 a); Auto.
    Inversion_clear s1_bst;  Apply H5; Auto.
    Apply X.eq_trans with t4; Auto.
    Elim (!ME.lt_not_gt a t1); Auto.
    Inversion_clear s2_bst.  
    Apply ME.lt_eq with t4; Auto.
    Inversion_clear s1_bst; Auto.
    (* ~(Subset t0 t3) *)
    Clear H; Right; Intuition.
    Apply f; Intros.
    Assert (In_tree a (Node t3 t4 t5 z0)).
    Apply H; Auto.
    Inversion_clear H1; Auto.
    Elim (!X.lt_not_eq a t1); Auto.
    Inversion_clear s1_bst; Auto.
    Apply X.eq_trans with t4; Auto.
    Elim (!ME.lt_not_gt a t1); Auto.
    Inversion_clear s1_bst; Auto.
    Inversion_clear s2_bst.  
    Apply ME.eq_lt with t4; Auto.
    (* t4 < t1 *)
    Case (H (Node Leaf t1 t2 0) t5); Auto; Intros.
    Simpl; Omega.
    Constructor; Inversion_clear s1_bst; Auto.
    Inversion_clear s2_bst; Auto.
    (* Subset (Node Leaf t1 t2) t5 *)
    Intros; Case (H t0 (Node t3 t4 t5 z0)); Auto; Intros.
    Simpl; Omega.
    Inversion_clear s1_bst; Auto.
    (* Subset t0 (Node t3 t4 t5 z0) *)
    Clear H; Left; Intuition.
    Generalize (i a) (i0 a); Clear i i0; Inversion_clear H; Intuition.
    (* ~ (Subset t2 (Node t3 t4 t5 z0)) *)
    Clear H; Right; Intuition.
    Apply f; Intuition.
    Assert (In_tree a (Node t3 t4 t5 z0)).
    Apply H; Inversion_clear H0; Auto.
    Inversion_clear H1.
    Inversion_clear H1; Auto.
    Inversion_clear H0; Auto.
    Elim (!X.lt_not_eq t4 t1); Auto.
    Apply X.eq_trans with a; Auto.
    Inversion_clear H1.
    Elim (!ME.lt_not_gt a t1); Auto.
    Apply ME.eq_lt with t4; Auto.
    Inversion_clear s1_bst; Auto.
    Inversion_clear H0; Auto.
    Elim (!ME.lt_not_gt a t4); Auto.
    Inversion_clear s2_bst; Auto.
    Apply ME.lt_eq with t1; Auto.
    Inversion_clear H1.
    Elim (!ME.lt_not_gt a t1); Auto.
    Apply X.lt_trans with t4; Inversion_clear s2_bst; Auto.
    Inversion_clear s1_bst; Auto.
  Defined.

  (** [for_all] and [exists] *)

  Definition for_all : (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})(s:t)
     { (compat_P E.eq P) -> (For_all P s) } + 
     { (compat_P E.eq P) -> ~(For_all P s) }.
  Proof.
     Intros P Pdec (s,s_bst,s_avl); Unfold For_all In; Simpl. 
     Clear s_bst s_avl; Induction s; Simpl.
     (* Leaf *)
     Left; Intros; Inversion_clear H0.
     (* Node s1 t0 s0 z *)
     Case (Pdec t0); Intro.
     (* P t0 *)
     Case (Hrecs1); Clear Hrecs1; Intro.
     Case (Hrecs0); Clear Hrecs0; [Left|Right]; Intuition.
     Inversion_clear H2; Ground.
     Clear Hrecs0; Right; Ground.
     (* ~(P t0) *)
     Clear Hrecs0 Hrecs1; Right; Intros; Ground.
  Defined.

  Definition exists : (P:elt->Prop)(Pdec:(x:elt){P x}+{~(P x)})(s:t)
    { (compat_P E.eq P) -> (Exists P s) } + 
    { (compat_P E.eq P) -> ~(Exists P s) }.
  Proof.
    Intros P Pdec (s,s_bst,s_avl); Unfold Exists In; Simpl. 
    Clear s_bst s_avl; Induction s; Simpl.
    (* Leaf *)
    Right; Intros; Intro; Elim H0; Intuition; Inversion_clear H2.
    (* Node s1 t0 s0 z *)
    Case (Pdec t0); Intro.
    (* P t0 *)
    Clear Hrecs0 Hrecs1; Left; Intro; Exists t0; Auto.
    (* ~(P t0) *)
    Case Hrecs1; Clear Hrecs1; Intro.
    Left; Intro; Elim (e H); Intuition; Exists x; Intuition.
    Case Hrecs0; Clear Hrecs0; Intro.
    Left; Intro; Elim (e H); Intuition; Exists x; Intuition.
    Right; Intros; Intro.
    Elim H0; Intuition.
    Inversion_clear H4; Ground.
  Defined.

  (** * Fold *)

  Module L := Raw(E).

  Fixpoint fold_tree [A:Set; f:elt->A->A; s:tree] : A -> A :=
    Cases s of
    | Leaf => [a]a
    | (Node l x r _) => [a](fold_tree A f l (f x (fold_tree A f r a)))
    end.
  Implicits fold_tree [1].

  Definition fold_tree' := 
   [A:Set; f:elt->A->A; s:tree] (L.fold f (elements_tree s)).
  Implicits fold_tree' [1].

  Lemma fold_tree_equiv_aux : 
     (A:Set)(s:tree)(f:elt->A->A)(a:A; acc : (list elt))
     (L.fold f (elements_tree_aux acc s) a)
     = (fold_tree f s (L.fold f acc a)).
  Proof.
  Induction s.
  Simpl; Intuition.
  Simpl; Intros.
  Rewrite H.
  Rewrite <- H0.
  Simpl; Auto.
  Qed.

  Lemma fold_tree_equiv : 
     (A:Set)(s:tree)(f:elt->A->A; a:A)
     (fold_tree f s a)=(fold_tree' f s a).
  Proof.
  Unfold fold_tree' elements_tree. 
  Induction s; Simpl; Auto; Intros.
  Rewrite fold_tree_equiv_aux.
  Rewrite H0.
  Simpl; Auto.
  Qed.

  Definition fold : 
   (A:Set)(f:elt->A->A)(s:t)(i:A) 
   { r : A | (EX l:(list elt) | 
                  (Unique E.eq l) /\
                  ((x:elt)(In x s)<->(InList E.eq x l)) /\ 
                  r = (fold_right f i l)) }.
  Proof.
    Intros A f s i; Exists (fold_tree f s i).
    Rewrite fold_tree_equiv.
    Unfold fold_tree'.
    Elim (L.fold_1 (elements_tree_sort ? (is_bst s)) i f); Intro l.
    Exists l; Elim H; Clear H; Intros H1 (H2,H3); Split; Auto.
    Split; Auto.
    Intros x; Generalize (elements_tree_1 s x) (elements_tree_2 s x).
    Generalize (H2 x); Unfold In; Ground.
  Defined.

  (** * Comparison *)

  (** ** Relations [eq] and [lt] over trees *)
  
  Definition eq : t -> t -> Prop := Equal.

  Lemma eq_refl: (s:t)(eq s s). 
  Proof.
    Unfold eq Equal; Intuition.
  Save.

  Lemma eq_sym: (s,s':t)(eq s s') -> (eq s' s).
  Proof.
    Unfold eq Equal; Ground.
  Save.

  Lemma eq_trans: (s,s',s'':t)(eq s s') -> (eq s' s'') -> (eq s s'').
  Proof.
    Unfold eq Equal; Ground.
  Save.

  Lemma eq_L_eq : (s,s':t)
    (eq s s') -> (L.eq (elements_tree s) (elements_tree s')).
  Proof.
    Unfold eq Equal L.eq L.Equal; Intros.
    Generalize (elements_tree_1 s a) (elements_tree_2 s a)
               (elements_tree_1 s' a) (elements_tree_2 s' a) (H a).
    Intuition.
  Save.

  Lemma L_eq_eq : (s,s':t)
    (L.eq (elements_tree s) (elements_tree s')) -> (eq s s').
  Proof.
    Unfold eq Equal L.eq L.Equal; Intros.
    Generalize (elements_tree_1 s a) (elements_tree_2 s a)
               (elements_tree_1 s' a) (elements_tree_2 s' a) (H a).
    Intuition.
  Save.
  Hints Resolve eq_L_eq L_eq_eq.

  Definition lt : t -> t -> Prop := 
    [s1,s2:t](L.lt (elements_tree s1) (elements_tree s2)).

  Definition lt_trans : (s,s',s'':t)(lt s s') -> (lt s' s'') -> (lt s s'')
    := [s,s',s'':t][h][h'](L.lt_trans h h').

  Lemma lt_not_eq : (s,s':t)(lt s s') -> ~(eq s s').
  Proof.
    Unfold lt; Intros; Intro.
    Apply L.lt_not_eq with s:=(elements_tree s) s':=(elements_tree s'); Auto.
  Save.

  (** ** Comparison algorithm 

      The code for [compare s1 s2] is basically comparing
      [elements s1] and [elements s2] in lexicographic order. But it builds
      these lists lazily (it is doing deforestation). It uses
      a function [compare_aux] comparing two lists of trees.

      We slightly modify the original code to get a 
      simplest termination argument. 
      The resulting code is even a bit more efficient.
<<
    let rec compare_aux l1 l2 =
        match (l1, l2) with
        ([], []) -> 0
      | ([], _)  -> -1
      | (_, []) -> 1
      | (Empty :: t1, Empty :: t2) ->
          compare_aux t1 t2
      | (Node(Empty, v1, r1, _) :: t1, Node(Empty, v2, r2, _) :: t2) ->
          let c = Ord.compare v1 v2 in
          if c <> 0 then c else compare_aux (r1::t1) (r2::t2)
      | (Node(Empty, v1, r1, _) :: t1, Empty :: t2) ->
          compare_aux (Node(Empty, v1, r1, 0) :: t1) t2
      | (Node(l1, v1, r1, _) :: t1, t2) when l1 <> Empty ->
          compare_aux (l1 :: Node(Empty, v1, r1, 0) :: t1) t2
      | (Empty :: t1, Node(Empty, v2, r2, _) :: t2) ->
          compare_aux t1 (Node(Empty, v2, r2, 0) :: t2)
      | (t1, Node(l2, v2, r2, _) :: t2) (* when l2 <> Empty *) ->
          compare_aux t1 (l2 :: Node(Empty, v2, r2, 0) :: t2)
>>
*)

  (** ** Lists of trees *)

  (** [flatten l] returns the list of elements of [l] i.e. the list
      of elements actually compared *)
  
  Fixpoint flatten [l:(list tree)] : (list elt) :=
     Cases l of 
     | nil => []
     | (cons t r) => (app (elements_tree t) (flatten r)) end.

  (** [sorted_l l] expresses that elements in the trees of [l] are
      sorted, and that all trees in [l] are binary search trees. *)

  Inductive In_tree_l : elt -> (list tree) -> Prop :=
    | InHd : (x:elt)(s:tree)(l:(list tree))(In_tree x s) ->
             (In_tree_l x (cons s l))
    | InTl : (x:elt)(s:tree)(l:(list tree))(In_tree_l x l) ->
             (In_tree_l x (cons s l)).

  Hint In_tree_l := Constructors In_tree_l.

  Inductive sorted_l : (list tree) -> Prop :=
    | SortedLNil  : 
        (sorted_l [])
    | SortedLCons : 
        (s:tree)(l:(list tree))
        (bst s) -> (sorted_l l) ->
        ((x:elt)(In_tree x s) -> (y:elt)(In_tree_l y l) -> (X.lt x y)) ->
        (sorted_l (cons s l)).

  Hint sorted_l := Constructors sorted_l.

  Lemma sort_app : 
    (l1,l2:(list elt))(sort E.lt l1) -> (sort E.lt l2) ->
    ((x,y:elt)(InList E.eq x l1) -> (InList E.eq y l2) -> (X.lt x y)) ->
    (sort E.lt (app l1 l2)).
  Proof.
    Induction l1; Simpl; Intuition.
    Apply cons_sort; Auto.
    Apply H; Auto.
    Inversion_clear H0; Trivial.
    Induction l; Simpl; Intuition.
    Induction l2; Simpl; Intuition. 
    Apply cons_leA; Apply H2; Auto.
    Apply cons_leA; Auto.
    Inversion_clear H0; Inversion_clear H4; Auto.
  Save.

  Lemma in_app :
    (x:elt)(l1,l2:(list elt))
    (InList E.eq x (app l1 l2)) -> 
    ((InList E.eq x l1) \/ (InList E.eq x l2)).
  Proof.
    Induction l1; Simpl; Intuition.
    Inversion_clear H0; Auto.
    Elim (H l2 H1); Auto.
  Save.

  Lemma in_flatten :
    (x:elt)(l:(list tree))
    (InList E.eq x (flatten l)) -> (In_tree_l x l).
  Proof.
    Induction l; Simpl; Intuition.
    Inversion_clear H.
    Elim (in_app x ? ? H0); Auto.
  Save.

  Lemma sorted_flatten : 
    (l:(list tree))(sorted_l l) -> (sort E.lt (flatten l)).
  Proof.
    Induction l; Simpl; Intuition.
    Apply sort_app; Inversion H0; Auto.
    Intros; Apply H5; Auto.
    Apply in_flatten; Auto.
  Save.

  (** ** Termination of [compare_aux]

     The measure is given by
     [m([]) = 0], [m(s :: l) = m(s) + m(l)],
     [m(Leaf) = 1], and [m(Node(l,_,r,_) = 1 + 3m(l) + m(r)] *)

  Fixpoint measure_t [s:tree] : Z :=
    Cases s of Leaf => 1
             | (Node l _ r _) => 1+3*(measure_t l)+(measure_t r) end.

  Fixpoint measure_l [l:(list tree)] : Z :=
    Cases l of nil => 0
             | (cons s l) => (measure_t s)+(measure_l l) end.

  Tactic Definition Measure_t := Unfold measure_t; Fold measure_t.
  Tactic Definition Measure_l := Unfold measure_l; Fold measure_l.

  Lemma measure_t_1 : (s:tree)(measure_t s) >= 1.
  Proof.
    Induction s.
    Simpl; Omega.
    Intros.
    Measure_t; Omega. (* BUG Simpl! *)
  Save.

  Tactic Definition Measure_t_1 s := Generalize (measure_t_1 s); Intro.

  Lemma measure_t_3 : (l,r:tree)(x:elt)(z:Z)
    (measure_t (Node l x r z)) >= 3.
  Proof.
    Intros; Measure_t.
    Generalize (measure_t_1 l) (measure_t_1 r); Intros; Omega.
  Save.

  Tactic Definition Measure_t_3 l x r z := Generalize (measure_t_3 l x r z); Intro.

  Lemma measure_l_0 : (l:(list tree))(measure_l l)>=0.
  Proof.
    Induction l.
    Simpl; Omega.
    Intros.
    Measure_l; Measure_t_1 a; Omega.
  Save.

  Tactic Definition Measure_l_0 l := Generalize (measure_l_0 l); Intro.

  (** Induction principle over the sum of the measures for two lists *)

  Definition compare_rec2 :
    (P:(list tree)->(list tree)->Set)
    ((x,x':(list tree))
       ((y,y':(list tree))
            (measure_l y)+(measure_l y') < (measure_l x)+(measure_l x')
            ->(P y y')) 
       -> (P x x')) ->
    (x,x':(list tree))(P x x').
  Proof.
    Intros P H x x'.
    Apply well_founded_induction_type_2 
      with R:=[yy',xx':(list tree)*(list tree)]
              (measure_l (Fst yy'))+(measure_l (Snd yy')) < 
              (measure_l (Fst xx'))+(measure_l (Snd xx'));
    Auto.                      
    Apply Wf_nat.well_founded_lt_compat with
      f:=[xx':(list tree)*(list tree)]
         (absolu ((measure_l (Fst xx'))+(measure_l (Snd xx')))).
    Intros; Apply absolu_lt.
    Measure_l_0 '(Fst x0); Measure_l_0 '(Snd x0);
    Measure_l_0 '(Fst y); Measure_l_0 '(Snd y);
    Intros; Omega.
(***
    Intros P H.
    Assert (n:Z)0<=n->(x,x':(list tree))n=(measure_l x)+(measure_l x')->(P x x').
    Intros n Hn; Pattern n; Apply Z_lt_rec; Auto.
    Intuition.
    Apply H; Intros.
    Apply (H0 ((measure_l y)+(measure_l y'))).
    Measure_l_0 y; Measure_l_0 y'; Omega.
    Omega.
    Intros; Apply (H0 ((measure_l x)+(measure_l x'))); Trivial.
    Measure_l_0 x; Measure_l_0 x'; Omega.
***)
  Save.

  (** ** Lemmas for the correctness of [compare] *)

  Lemma lt_nil_elements_tree_Node :
    (l,r:tree)(x:elt)(z:Z)(L.lt [] (elements_tree (Node l x r z))).
  Proof.
    Unfold elements_tree; Simpl; Intros l r x z; Clear z.
    Generalize (elements_tree_aux [] r) x.
    NewInduction l; Simpl; Intuition.
  Save.

  Lemma lt_app : (l1,l2,l3:(list elt))
    (L.lt l1 l2) -> (L.lt l1 (app l2 l3)). 
  Proof.
    Induction 1; Simpl; Intuition.
  Save.

  Lemma lt_app_eq : (l1,l2,l3:(list elt))
    (L.lt l2 l3) -> (L.lt (app l1 l2) (app l1 l3)). 
  Proof.
    Induction l1; Simpl; Intuition.
  Save.

  Lemma lt_eq_1 : (l1,l2,l3:(list elt))
    l1=l2 -> (L.lt l1 l3) -> (L.lt l2 l3). 
  Proof.
    Intros; Rewrite <- H; Auto.
  Save.

  Lemma lt_eq_2 : (l1,l2,l3:(list elt))
    l2=l3 -> (L.lt l1 l2) -> (L.lt l1 l3). 
  Proof.
    Intros; Rewrite <- H; Auto.
  Save.

  Lemma eq_eq_1 : (l1,l2,l3:(list elt))
    l1=l2 -> (L.eq l1 l3) -> (L.eq l2 l3). 
  Proof.
    Intros; Rewrite <- H; Auto.
  Save.

  Lemma eq_eq_2 : (l1,l2,l3:(list elt))
    l2=l3 -> (L.eq l1 l2) -> (L.eq l1 l3). 
  Proof.
    Intros; Rewrite <- H; Auto.
  Save.

  Lemma l_eq_cons : (l1,l2:(list elt))(x,y:elt)
    (X.eq x y) -> (L.eq l1 l2) -> (L.eq (x :: l1) (y :: l2)).
  Proof.
    Unfold L.eq L.Equal; Intuition.
    Inversion_clear H1; Generalize (H0 a); Clear H0; Intuition.
    Apply L.In_eq with x; Auto.
    Inversion_clear H1; Generalize (H0 a); Clear H0; Intuition.
    Apply L.In_eq with y; Auto.
  Save.

  Hints Resolve lt_nil_elements_tree_Node lt_app lt_app_eq 
                lt_eq_1 lt_eq_2 eq_eq_1 eq_eq_2 l_eq_cons.

  Lemma elements_app : (s:tree)(acc:(list elt))
    (elements_tree_aux acc s) = (app (elements_tree s) acc).
  Proof.
    Induction s; Simpl; Intuition.
    Rewrite H0.
    Rewrite H.
    Unfold elements_tree; Simpl.
    Do 2 Rewrite H.
    Rewrite H0.
    Repeat Rewrite <- app_nil_end.
    Repeat Rewrite app_ass; Auto.
  Save.

  (** main lemma for correctness of [compare] *)
  Lemma compare_flatten :
    (l,r:tree)(x:elt)(z:Z)(tl:(list tree))
    (flatten ((Node l x r z) :: tl)) 
  = (flatten (l :: ((Node Leaf x r z) :: tl))).
  Proof.
    Simpl; Unfold elements_tree; Simpl; Intuition.
    Repeat Rewrite elements_app.
    Repeat Rewrite <- app_nil_end.
    Repeat Rewrite app_ass; Auto.
  Save.

  Hints Resolve compare_flatten.

  (** same lemma, expressed differently *)
  Lemma compare_flatten_1 :
    (t0,t2:tree)(t1:elt)(z:Z)(l:(list elt))
    (app (elements_tree t0) (t1 :: (app (elements_tree t2) l)))
  = (app (elements_tree (Node t0 t1 t2 z)) l).
  Proof.
    Simpl; Unfold elements_tree; Simpl; Intuition.
    Repeat Rewrite elements_app.
    Repeat Rewrite <- app_nil_end.
    Repeat Rewrite app_ass; Auto.
  Save.

  Hints Resolve compare_flatten_1.

  (** invariant for [compare l1 l2]: [Leaf] may only occur on head
      of [l1] and [l2], and only when the other list is non-empty *)

  Fixpoint no_leaf [l:(list tree)] : Prop := Cases l of 
    | nil => True
    | (cons Leaf _) => False
    | (cons _ r) => (no_leaf r) end.

  Inductive leaf_invariant : (list tree) -> (list tree) -> Prop := 
  | LI_nil_l : (l:(list tree)) (no_leaf l) -> (leaf_invariant [] l)
  | LI_l_nil : (l:(list tree)) (no_leaf l) -> (leaf_invariant l [])
  | LI_leaf_leaf : (l1,l2:(list tree))(no_leaf l1) -> (no_leaf l2) ->
                   (leaf_invariant (Leaf :: l1) (Leaf :: l2))
  | LI_leaf_l : (l1,l2:(list tree))(no_leaf l1) -> (no_leaf l2) ->
                ~l2=[] -> (leaf_invariant (Leaf :: l1) l2)
  | LI_l_leaf : (l1,l2:(list tree))(no_leaf l1) -> (no_leaf l2) ->
                ~l1=[] -> (leaf_invariant l1 (Leaf :: l2))
  | LI_l_l : (l1,l2:(list tree))(no_leaf l1) -> (no_leaf l2) ->
                ~l1=[] -> ~l2=[] -> (leaf_invariant l1 l2).

  Hint leaf_invariant := Constructors leaf_invariant.

  Lemma no_leaf_invariant : (l1,l2:(list tree))
    (no_leaf l1) -> (no_leaf l2) -> (leaf_invariant l1 l2).
  Proof.
    Destruct l1; Destruct l2; Intuition.
  Save.

  Hints Resolve no_leaf_invariant.

  (** ** [compare_aux] and [compare] *) 

  Definition compare_aux :
    (l1,l2:(list tree))(sorted_l l1) -> (sorted_l l2) ->
    (leaf_invariant l1 l2) -> 
    (Compare L.lt L.eq (flatten l1) (flatten l2)).
  Proof.
    Intros l1 l2; Pattern l1 l2; Apply compare_rec2.
    Destruct x; Destruct x'; Intuition.
    (* x = x' = nil *)
    Constructor 2; Unfold L.eq L.Equal; Intuition.
    (* x = nil *)
    Constructor 1; Simpl; Auto.
    Inversion_clear H2; Intuition.
    NewDestruct t0; Inversion_clear H1; Simpl.
    Elim H3.
    Auto.
    (* x <> nil, x' = nil *)
    Constructor 3; Simpl; Auto.
    NewDestruct t0; Inversion_clear H0; Simpl.
    Inversion_clear H2; Intuition.
    Elim H0.
    Auto.
    (* x,x' <> nil *)
    NewDestruct t0; NewDestruct t1; Simpl.
    (* Leaf :: l, Leaf :: l0 *)
    Case (H l l0); Clear H; Auto; Intros.
    Measure_l; Unfold measure_t; Omega.
    Inversion_clear H0; Trivial.
    Inversion_clear H1; Trivial.
    Inversion_clear H2; Intuition.
    Elim H3. Elim H. Elim H.
    Constructor 1; Auto.
    Constructor 2; Auto.    
    Constructor 3; Auto.
    (* Leaf :: l, (Node t0 t1 t2) :: l0 *)
    NewDestruct t0; Simpl.
    (* Leaf :: l, (Node Leaf t1 t2) :: l0 *)
    Case (H l (Node Leaf t1 t2 z) :: l0); Clear H; Auto; Intros.
    Measure_l; Measure_t; Omega.
    Inversion_clear H0; Auto.
    Inversion_clear H2; Intuition; Elim H.
    Constructor 1; Auto.
    Constructor 2; Auto.
    Constructor 3; Auto.
    (* Leaf :: l, (Node (Node t0 t3 t4) t1 t2) :: l0 *)
    Case (H (Leaf::l) (Node t0 t3 t4 z0) :: ((Node Leaf t1 t2 z) :: l0)); Clear H; Auto; Intros.
    Measure_l; Measure_t. Measure_t_1 t0; Measure_t_1 t4; Omega.
    Constructor; Inversion_clear H1; Auto.
    Inversion_clear H; Auto.
    Constructor; Auto.
    Inversion_clear H; Auto.
    Intros; Inversion_clear H1.
    Apply H4; Auto.
    Inversion_clear H6.
    Apply H4; Auto.
    Intros; Inversion_clear H5; Auto.
    Inversion_clear H.
    Inversion_clear H6; Auto.
    Apply ME.lt_eq with t1; Auto.
    Inversion_clear H.
    Apply X.lt_trans with t1; Auto.
    Inversion_clear H2; Intuition.
    Constructor 1; Rewrite <- compare_flatten_1; Auto.
    Constructor 2; Rewrite <- compare_flatten_1; Auto.
    Constructor 3; Rewrite <- compare_flatten_1; Auto.
    (* (Node t0 t2 t3) :: l, Leaf :: l0 *)
    NewDestruct t0; Simpl.
    (* (Node Leaf t2 t3) :: l, Leaf :: l0 *)
    Case (H ((Node Leaf t2 t3 z) :: l) l0); Clear H; Auto; Intros.
    Measure_l; Measure_t; Omega.
    Inversion_clear H1; Auto.
    Inversion_clear H2; Intuition; Elim H3.
    Constructor 1; Auto.
    Constructor 2; Auto.
    Constructor 3; Auto.
    (* (Node (Node t0 t1 t4) t2 t3) :: l, Leaf :: l0 *)
    Case (H ((Node t0 t1 t4 z0) :: ((Node Leaf t2 t3 z) :: l)) (Leaf::l0)); Clear H; Auto; Intros.
    Measure_l; Measure_t. Measure_t_1 t0; Measure_t_1 t4; Omega.
    Constructor; Inversion_clear H0; Auto.
    Inversion_clear H; Auto.
    Constructor; Auto.
    Inversion_clear H; Auto.
    Intros; Inversion_clear H0.
    Apply H4; Auto.
    Inversion_clear H6.
    Apply H4; Auto.
    Intros; Inversion_clear H5; Auto.
    Inversion_clear H6; Auto.
    Apply ME.lt_eq with t2; Auto.
    Inversion_clear H; Auto.
    Inversion_clear H5.
    Apply X.lt_trans with t2; Inversion_clear H; Auto.
    Inversion_clear H2; Intuition; Elim H3.
    Constructor 1; Rewrite <- compare_flatten_1; Auto. 
    Constructor 2; Rewrite <- compare_flatten_1; Auto. 
    Constructor 3; Rewrite <- compare_flatten_1; Auto. 
    (* (Node t0 t2 t3)::l, (Node t1 t4 t5)::l0) *)
    NewDestruct t0; NewDestruct t1; Simpl.
    (* (Node Leaf t2 t3)::l, (Node Leaf t4 t5)::l0) *)
    Case (X.compare t2 t4); Intro.
    (* t2 < t4 *)
    Constructor 1; Auto.
    (* t2 = t4 *)
    Case (H (t3::l) (t5::l0)); Clear H; Auto; Intros.
    Measure_l; Measure_t. Measure_t_1 t3; Measure_t_1 t5; Omega.
    Constructor; Inversion_clear H0; Auto.
    Inversion_clear H; Trivial.
    Inversion_clear H1; Constructor; Intuition.
    Inversion_clear H; Trivial.
    Inversion_clear H2; Intros.
    NewDestruct t3; NewDestruct t5; Intuition.
    Constructor 1; Auto.
    Constructor 2; Auto.
    Constructor 3; Auto.
    (* t4 < t2 *)
    Constructor 3; Auto.
    (* (Node Leaf t2 t3)::l, (Node (Node t0 t1 t6) t4 t5)::l0) *)
    Case (H ((Node Leaf t2 t3 z)::l) 
            ((Node t0 t1 t6 z1)::((Node Leaf t4 t5 z0)::l0)));
    Clear H; Auto; Intros.
    Measure_l; Measure_t. 
    Measure_t_1 t3; Measure_t_1 t5; Measure_t_1 t0; Measure_t_1 t6; Omega.
    Inversion_clear H1.
    Constructor; Intuition.
    Inversion_clear H; Trivial.
    Constructor; Intuition.
    Inversion_clear H; Auto.
    Inversion_clear H1; Auto.
    Inversion_clear H6.
    Inversion_clear H.
    Inversion_clear H5; Auto.
    Inversion_clear H; Auto.
    Apply ME.lt_eq with t4; Auto.
    Inversion_clear H5.
    Apply X.lt_trans with t4; Auto.
    Inversion_clear H2; Intuition.
    Constructor 1; Rewrite <- compare_flatten_1; Auto. 
    Constructor 2; Rewrite <- compare_flatten_1; Auto. 
    Constructor 3; Rewrite <- compare_flatten_1; Auto. 
    (* Node (Node t0 t6 t7) t2 t3 :: l, Node Leaf t4 t5 :: l0 *)
    Case (H ((Node t0 t6 t7 z1)::((Node Leaf t2 t3 z)::l))
            ((Node Leaf t4 t5 z0)::l0));
    Clear H; Auto; Intros.
    Measure_l; Measure_t. 
    Measure_t_1 t3; Measure_t_1 t5; 
    Measure_t_1 t0; Measure_t_1 t5; Measure_t_1 t7; Omega.
    Inversion_clear H0.
    Constructor; Intuition.
    Inversion_clear H; Trivial.
    Constructor; Intuition.
    Inversion_clear H; Auto.
    Inversion_clear H0; Auto.
    Inversion_clear H6.
    Inversion_clear H.
    Inversion_clear H5; Auto.
    Inversion_clear H; Auto.
    Apply ME.lt_eq with t2; Auto.
    Inversion_clear H5.
    Apply X.lt_trans with t2; Auto.
    Inversion_clear H2; Intuition.
    Constructor 1; Rewrite <- compare_flatten_1; Auto. 
    Constructor 2; Rewrite <- compare_flatten_1; Auto. 
    Constructor 3; Rewrite <- compare_flatten_1; Auto. 
    (* Node (Node t0 t6 t7) t2 t3 :: l, Node (Node t1 t8 t9) t4 t5 :: l0 *)
    Case (H ((Node t0 t6 t7 z1)::((Node Leaf t2 t3 z)::l))
            ((Node (Node t1 t8 t9 z2) t4 t5 z0)::l0));
    Clear H; Auto; Intros.
    Measure_l; Measure_t. 
    Measure_t_1 t3; Measure_t_1 t5; 
    Measure_t_1 t0; Measure_t_1 t5; Measure_t_1 t7; Omega.
    Inversion_clear H0.
    Constructor; Intuition.
    Inversion_clear H; Trivial.
    Constructor; Intuition.
    Inversion_clear H; Auto.
    Inversion_clear H0; Auto.
    Inversion_clear H6.
    Inversion_clear H.
    Inversion_clear H5; Auto.
    Inversion_clear H; Auto.
    Apply ME.lt_eq with t2; Auto.
    Inversion_clear H5.
    Apply X.lt_trans with t2; Auto.
    Inversion_clear H2; Intuition.
    Constructor 1; Rewrite <- compare_flatten_1; Auto. 
    Constructor 2; Rewrite <- compare_flatten_1; Auto. 
    Constructor 3; Rewrite <- (compare_flatten_1 (Node t0 t6 t7 z1)); Auto. 
  Defined.

  Lemma flatten_elements : 
    (s:tree)(flatten (s::[])) = (elements_tree s).
  Proof.
    Simpl; Intros; Rewrite <- app_nil_end; Auto.
  Save.

  Definition compare : (s1,s2:t)(Compare lt eq s1 s2).
  Proof.
    Intros (s1,s1_bst,s1_avl) (s2,s2_bst,s2_avl); Unfold lt eq; Simpl.
    Case (compare_aux (s1 :: []) (s2 :: [])); Intros.
    Constructor; Intuition; Inversion_clear H0.
    Constructor; Intuition; Inversion_clear H0.
    NewDestruct s1; NewDestruct s2; Simpl; Constructor; Simpl; Intuition;
    Discriminate H.
    Constructor 1; Simpl; Repeat Rewrite <- flatten_elements; Auto.
    Constructor 2.
    Apply L_eq_eq; Simpl; Repeat Rewrite <- flatten_elements; Auto.
    Constructor 3; Simpl; Repeat Rewrite <- flatten_elements; Auto.
  Defined.

  (** * Equality test *)

  Definition equal : (s,s':t){ Equal s s' } + { ~(Equal s s') }.
  Proof.
    Intros s s'; Case (compare s s'); Intro.
    Right; Apply lt_not_eq; Auto.
    Left; Auto.
    Right; Intro; Apply (lt_not_eq s' s); Auto; Apply eq_sym; Auto.
  Defined.

  Write State toto.

End Make.







