(***********************************************************************)
(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
(*   \VV/  *************************************************************)
(*    //   *      This file is distributed under the terms of the      *)
(*         *       GNU Lesser General Public License Version 2.1       *)
(***********************************************************************)

(* Finite sets library.  
 * Authors: Pierre Letouzey and Jean-Christophe Filliâtre 
 * Institution: LRI, CNRS UMR 8623 - Université Paris Sud
 *              91405 Orsay, France *)

(* $Id: FSetProperties.v,v 1.10 2003/07/18 15:33:07 filliatr Exp $ *)

(** This functor derives additional properties from [FSetInterface.S].
    Contrary to the functor in [FSetProperties] it uses 
    predicates over sets instead of sets operations, i.e.
    [In x s] instead of [mem x s=true], 
    [Equal s s'] instead of [equal s s'=true], etc. *)

Require Omega.

Import nat_scope.
Open Scope nat_scope.

Require Export FSetInterface. 
Require Export Bool.
Require Export Sumbool.
Require Export Zerob. 
Set Implicit Arguments.

Section Misc.
Variable A,B : Set.
Variable eqA : A -> A -> Prop. 
Variable eqB : B -> B -> Prop.

(** Two-argument functions that allow to reorder its arguments. *)
Definition transpose := [f:A->B->B](x,y:A)(z:B)(eqB (f x (f y z)) (f y (f x z))). 

(** Compatibility of a two-argument function with respect to two equalities. *)
Definition compat_op := [f:A->B->B](x,x':A)(y,y':B)(eqA x x') -> (eqB y y') -> 
 (eqB (f x y) (f x' y')).

(** Compatibility of a function upon natural numbers. *)
Definition compat_nat := [f:A->nat] (x,x':A)(eqA x x') -> (f x)=(f x').

End Misc.
Hints Unfold transpose compat_op compat_nat.

(** To prove [(Setoid_Theory ? (eq ?))] *)
Tactic Definition ST :=
  Constructor; Intros;[ Trivial | Symmetry; Trivial | EApply trans_eq; EAuto ].
Hint st : core := Extern 1 (Setoid_Theory ? (eq ?)) ST.

Definition gen_st : (A:Set)(Setoid_Theory ? (eq A)).
Auto.
Qed.

(** Usual syntax for lists. 
    CAVEAT: the Coq cast "::" will no longer be available. *)
Notation "[]" := (nil ?) (at level 1).
Notation "a :: l" := (cons a l) (at level 1, l at next level).

Module Properties [M:S].
  Import M.
  Import Logic. (* to unmask [eq] *)  
  Import Peano. (* to unmask [lt] *)
  
  Module ME := MoreOrderedType E.  

  (** * Alternative (weaker) specification for [fold],
        based on [empty] ([fold_1]) and [add] ([fold_2]). *)

  Section Old_Spec_Now_Properties. 

   (** Results about lists without duplicates *)

   Section Unique_Remove.

   Fixpoint remove_list [x:elt;s:(list elt)] : (list elt) := Cases s of 
      nil => []
    | (cons y l) => if (ME.eq_dec x y) then [_]l else [_]y::(remove_list x l)
   end. 

   Lemma remove_list_correct : 
    (s:(list elt))(x:elt)(Unique E.eq s) -> 
    (Unique E.eq (remove_list x s)) /\ 
    ((y:elt)(InList E.eq y (remove_list x s))<->(InList E.eq y s)/\~(E.eq x y)).
   Proof.
   Induction s; Simpl.
   Split; Auto.
   Intuition.
   Inversion H0.
   Intros; Inversion_clear H0; Case (ME.eq_dec x a); Trivial.  
   Intuition.
   Apply H1; EApply ME.In_eq with y; EAuto.
   Inversion_clear H3; Auto.
   Elim H4; EAuto. 
   Elim (H x H2); Intros.
   Split.
   Elim (H3 a); Constructor; Intuition.
   Intro y; Elim (H3 y); Clear H3; Intros.
   Intuition.  
   Inversion_clear H4; Auto.
   Elim (H3 H6); Auto.
   Inversion_clear H4; Auto.
   Intuition EAuto.
   Elim (H3 H7); Ground.
   Inversion_clear H6; Ground. 
   Qed. 

   Local ListEq := [l,l'](y:elt)(InList E.eq y l)<->(InList E.eq y l').
   Local ListAdd := [x,l,l'](y:elt)(InList E.eq y l')<->(E.eq y x)\/(InList E.eq y l).

   Lemma remove_list_equal : 
    (s,s':(list elt))(x:elt)(Unique E.eq x::s) -> (Unique E.eq s') -> 
    (ListEq x::s s') -> (ListEq s (remove_list x s')).
   Proof.  
   Unfold ListEq; Intros. 
   Inversion_clear H.
   Elim (remove_list_correct x H0); Intros.
   Elim (H4 y); Intros.
   Elim (H1 y); Intros.
   Split; Intros.
   Apply H6; Split; Auto. 
   Intro.
   Elim H2; Apply ME.In_eq with y; EAuto.
   Elim (H5 H9); Intros.
   Assert H12 := (H8 H10). 
   Inversion_clear H12; Auto.
   Elim H11; EAuto. 
   Qed. 

   Lemma remove_list_add : 
    (s,s':(list elt))(x,x':elt)(Unique E.eq s) -> (Unique E.eq x'::s') -> 
    ~(E.eq x x') -> ~(InList E.eq x s) -> 
    (ListAdd x s x'::s') -> (ListAdd x (remove_list x' s) s').
   Proof.
   Unfold ListAdd; Intros.
   Inversion_clear H0.
   Elim (remove_list_correct x' H); Intros.
   Elim (H6 y); Intros.
   Elim (H3 y); Intros.
   Split; Intros.
   Elim H9; Auto; Intros.
   Elim (ME.eq_dec y x); Auto; Intros.
   Right; Apply H8; Split; Auto.
   Intro; Elim H4; Apply ME.In_eq with y; Auto.
   Inversion_clear H11.
   Assert (InList E.eq y x'::s'). Auto.
   Inversion_clear H11; Auto.
   Elim H1; EAuto.
   Elim (H7 H12); Intros.
   Assert (InList E.eq y x'::s'). Auto.
   Inversion_clear H14; Auto.
   Elim H13; Auto. 
   Qed.

   Lemma remove_list_fold_right : 
    (A:Set)(eqA:A->A->Prop)(st:(Setoid_Theory A eqA))
    (i:A)(f:elt->A->A)(compat_op E.eq eqA f) -> (transpose eqA f) -> 
    (s:(list elt))(x:elt)(Unique E.eq s) -> (InList E.eq x s) -> 
    (eqA (fold_right f i s) (f x (fold_right f i (remove_list x s)))).
   Proof.
   Induction s; Simpl.  
   Intros; Inversion H2.
   Intros.
   Inversion_clear H2.
   Case  (ME.eq_dec x a); Simpl; Intros.
   Apply H; Auto. 
   Apply Seq_refl; Auto. 
   Inversion_clear H3. 
   Elim n; Auto.
   Apply (Seq_trans ?? st) with (f a (f x (fold_right f i (remove_list x l)))).
   Apply H; Auto.
   Apply H0; Auto.
   Qed.   

   Lemma fold_right_equal : 
    (A:Set)(eqA:A->A->Prop)(st:(Setoid_Theory A eqA))
    (i:A)(f:elt->A->A)(compat_op E.eq eqA f) -> (transpose eqA f) -> 
    (s,s':(list elt))(Unique E.eq s) -> (Unique E.eq s') -> (ListEq s s') -> 
    (eqA (fold_right f i s) (fold_right f i s')).
   Proof.
   Induction s.
   Intro s'; Case s'; Simpl. 
   Intros; Apply Seq_refl; Auto.
   Unfold ListEq; Intros.
   Elim (H3 e); Intros. 
   Assert X : (InList E.eq e []); Auto; Inversion X.
   Intros x l Hrec s' U U' E.
   Simpl.   
   Apply (Seq_trans ?? st) with (f x (fold_right f i (remove_list x s'))).
   Apply H; Auto.
   Apply Hrec; Auto.
   Inversion U; Auto.
   Elim (remove_list_correct x U'); Auto.
   Apply remove_list_equal; Auto.
   Apply Seq_sym; Auto.
   Apply remove_list_fold_right with eqA:=eqA; Auto.
   Unfold ListEq in E; Ground.
   Qed.

   Lemma fold_right_add : 
    (A:Set)(eqA:A->A->Prop)(st:(Setoid_Theory A eqA))
    (i:A)(f:elt->A->A)(compat_op E.eq eqA f) -> (transpose eqA f) -> 
    (s',s:(list elt))(x:elt)(Unique E.eq s) -> (Unique E.eq s') -> ~(InList E.eq x s) -> 
    (ListAdd x s s') -> 
    (eqA (fold_right f i s') (f x (fold_right f i s))).
   Proof.   
   Induction s'.
   Unfold ListAdd; Intros.
   Elim (H4 x); Intros. 
   Assert X : (InList E.eq x []); Auto; Inversion X.
   Intros x' l' Hrec s x U U' IN EQ; Simpl.
   (* if x=x' *)
   Case (ME.eq_dec x x'); Intros.
   Apply H; Auto.
   Apply fold_right_equal with eqA:=eqA; Auto.
   Inversion_clear U'; Trivial.
   Unfold ListEq; Unfold ListAdd in EQ.
   Intros. 
   Elim (EQ y); Intros.
   Split; Intros.
   Elim H1; Auto.
   Intros; Inversion_clear U'.
   Elim H5; Apply ME.In_eq with y; EAuto.
   Assert (InList E.eq y x'::l'); Auto; Inversion_clear H4; Auto.
   Elim IN; Apply ME.In_eq with y; EAuto.
   (* else x<>x' *)   
   Apply (Seq_trans ?? st) with (f x' (f x (fold_right f i (remove_list x' s)))).
   Apply H; Auto.
   Apply Hrec; Auto.
   Elim (remove_list_correct x' U); Auto.
   Inversion_clear U'; Auto.
   Elim (remove_list_correct x' U); Intros; Intro.
   Ground.
   Apply remove_list_add; Auto.
   Apply (Seq_trans ?? st) with (f x (f x' (fold_right f i (remove_list x' s)))).
   Apply H0; Auto.
   Apply H; Auto.
   Apply Seq_sym; Auto.
   Apply remove_list_fold_right with eqA:=eqA; Auto.
   Elim (EQ x'); Intros. 
   Elim H1; Auto; Intros; Elim n; Auto.
   Qed.

  End Unique_Remove.

  (** An alternate (and previous) specification for [fold] was based on 
      the recursive structure of a set. It is now lemmas [fold_1] and 
      [fold_2]. *)

  Lemma fold_1:
   (s:t)(A:Set)(eqA:A->A->Prop)(st:(Setoid_Theory A eqA))(i:A)(f:elt->A->A)
   (Empty s) -> (eqA (fold f s i) i).
  Proof.
  Intros; Elim (M.fold_1 s i f); Intros l (H1,(H2,H3)).
  Rewrite H3; Clear H3.
  Unfold Empty in H; Generalize H H2; Clear H H2; Case l; Simpl; Intros.
  Apply Seq_refl; Trivial.
  Elim (H e).
  Elim (H2 e); Intuition. 
  Qed.

  Lemma fold_2 : 
     (s,s':t)(x:elt)(A:Set)(eqA:A->A->Prop)(st:(Setoid_Theory A eqA))
     (i:A)(f:elt->A->A)(compat_op E.eq eqA f) -> (transpose eqA f) -> ~(In x s) -> 
     (Add x s s') -> (eqA (fold f s' i) (f x (fold f s i))).
  Proof.
  Intros; Elim (M.fold_1 s i f); Intros l (Hl,(Hl1,Hl2)).
  Elim (M.fold_1 s' i f); Intros l' (Hl',(Hl'1,Hl'2)).
  Rewrite Hl2; Clear Hl2.
  Rewrite Hl'2; Clear Hl'2.
  Assert (y:elt)(InList E.eq y l')<->(E.eq y x)\/(InList E.eq y l).
   Intros; Elim (H2 y); Intros; Split; 
    Elim (Hl1 y); Intros; Elim (Hl'1 y); Intuition.
  Assert ~(InList E.eq x l).
   Intro; Elim H1; Ground.
  Clear H1 H2 Hl'1 Hl1 H1 s' s.
  Apply fold_right_add with eqA:=eqA; Auto.
  Qed.

  (** Similar specification for [cardinal]. *)

  Lemma cardinal_fold : (s:t)(cardinal s)=(fold [_]S s O).
  Proof.
  Intros; Elim (M.cardinal_1 s); Intros l (Hl,(Hl1,Hl2)).
  Elim (M.fold_1 s O [_]S); Intros l' (Hl',(Hl'1,Hl'2)).
  Rewrite Hl2; Rewrite Hl'2; Clear Hl2 Hl'2.
  Assert (l:(list elt))(length l)=(fold_right [_]S O l).
   Induction l0; Simpl; Auto.
  Rewrite H.
  Apply fold_right_equal with eqA:=(eq nat); Auto; Ground.
  Qed.

  Lemma cardinal_1 : (s:t)(Empty s) -> (cardinal s)=O.
  Proof.
  Intros; Rewrite cardinal_fold; Apply fold_1; Auto.
  Qed.

  Lemma cardinal_2 : 
    (s,s':t)(x:elt)~(In x s) -> (Add  x s s') -> (cardinal s') = (S (cardinal s)).
  Proof.
  Intros; Do 2 Rewrite cardinal_fold.
  Change S with ([_]S x).
  Apply fold_2 with eqA:=(eq nat); Auto.
  Qed.

  Hints Resolve cardinal_1 cardinal_2.

  End Old_Spec_Now_Properties.


  (** * Induction principle over sets *)

  Lemma Add_add : 
    (s:t)(x:elt)(Add x s (add x s)).
  Proof.
    Unfold Add; Intros; Intuition.
    Elim (ME.eq_dec x y); Intros; Auto.
    Right.
    EApply add_3; EAuto.
  Qed.

  Lemma Add_remove : (s:t)(x:elt)(In x s) -> (Add x (remove x s) s). 
  Proof. 
    Intros; Unfold Add; Intuition.
    Elim (ME.eq_dec x y); Intros; Auto.
    Apply In_1 with x; Auto.
    EAuto.
  Qed.
 
  Hints Resolve Add_add Add_remove. 

  Lemma Equal_remove : (s,s':t)(x:elt)(In x s)->(Equal s s')-> 
     (Equal (remove x s) (remove x s')).
  Proof.
     Unfold Equal; Intros.
     Elim (ME.eq_dec x a); Intros; Auto. 
     Split; Intros.
     Absurd (In x (remove x s)); Auto; Apply In_1 with a; Auto.
     Absurd (In x (remove x s')); Auto; Apply In_1 with a; Auto. 
     Elim (H0 a); Intros. 
     Split; Intros; Apply remove_2; Auto; 
      [Apply H1|Apply H2]; EApply remove_3;EAuto.    
  Save.

  Hints Resolve Equal_remove.

  Lemma cardinal_inv_1 : (s:t)(cardinal s)=O -> (Empty s). 
  Proof. 
    Intros; Generalize (!choose_1 s) (!choose_2 s).
    Elim (choose s); Intuition.
    Clear H1; Assert (In a s); Auto.
    Rewrite (!cardinal_2 (remove a s) s a) in H; Auto.
    Inversion H.
  Save.
  Hints Resolve cardinal_inv_1.
 
  Lemma cardinal_inv_2 : 
     (s:t)(n:nat)(cardinal s)=(S n) -> (EX x:elt | (In x s)).
  Proof. 
    Intros; Generalize (!choose_1 s) (!choose_2 s).
    Elim (choose s); Intuition. 
    Exists a; Auto.
    Intros; Rewrite cardinal_1 in H; Auto; Inversion H.
  Qed.

  Lemma Equal_cardinal_aux : (n:nat)(s,s':t)(cardinal s)=n -> 
       (Equal s s')->(cardinal s)=(cardinal s').
  Proof.
     Induction n.
     Intros.  
     Rewrite H.
     Symmetry.
     Apply cardinal_1.
     Generalize (cardinal_inv_1 H) H0.
     Unfold Empty Equal; Intuition.
     Generalize (H1 a) (H2 a); Intuition.
     Intros.
     Elim (cardinal_inv_2 H0); Intros.
     Assert (In x s'). 
       Generalize (H1 x); Intuition.  
     Generalize H0.
     Rewrite (!cardinal_2 (remove x s) s x);Auto.
     Rewrite (!cardinal_2 (remove x s') s' x); Auto.
  Qed.

  Lemma Equal_cardinal : (s,s':t)(Equal s s')->(cardinal s)=(cardinal s').
  Proof. 
    Intros; EApply Equal_cardinal_aux; EAuto.
  Qed.

  Hints Resolve Add_add Add_remove Equal_remove 
                cardinal_inv_1 Equal_cardinal.

  Lemma empty_cardinal: (cardinal empty)=O.
  Proof.
  Rewrite cardinal_fold; Auto.
  Apply fold_1; Auto.
  Qed.

  Lemma empty_cardinal_2 : (s:t)(Empty s) -> (cardinal s)=O.
  Proof.
  Intros; Replace O with (cardinal empty). 
  Apply Equal_cardinal. 
  Unfold Empty in H; Unfold Equal; Ground.
  Absurd (In a empty); Auto.
  Apply empty_cardinal.
  Qed.

  Hints Immediate empty_cardinal empty_cardinal_2 : set.

  Lemma cardinal_induction : (P : t -> Prop)
     ((s:t)(Empty s)->(P s)) -> 
     ((s,s':t)(P s) -> (x:elt)~(In x s) -> (Add x s s') -> (P s')) -> 
     (n:nat)(s:t)(cardinal s)=n -> (P s).
  Proof.
  Induction n.
  Intros; Apply H; Auto.
  Intros; Elim (cardinal_inv_2 H2); Intros. 
  Apply H0 with (remove x s) x; Auto.
  Apply H1; Auto.
  Assert (S (cardinal (remove x s))) = (S n0); Auto. 
  Rewrite <- H2; Symmetry.
  EApply cardinal_2; EAuto.
  Qed.

  Lemma set_induction : (P : t -> Prop)
     ((s:t)(Empty s)->(P s)) -> 
     ((s,s':t)(P s) -> (x:elt)~(In x s) -> (Add x s s') -> (P s')) -> 
     (s:t)(P s).
  Proof.
  Intros; EApply cardinal_induction; EAuto.
  Qed.  

  (** Other formulation of [fold_1] and [fold_2]. *)

  Section Fold. 

  Variable s,s':t.
  Variable x:elt.

  Variable A:Set. 
  Variable eqA:A->A->Prop.
  Variable st:(Setoid_Theory ? eqA).
  Variable i:A.
  Variable f:elt->A->A.
  Variable Comp:(compat_op E.eq eqA f). 
  Variable Assoc:(transpose eqA f).
 
  Lemma fold_empty: (eqA (fold f empty i) i).
  Proof. 
  Apply fold_1; Auto.
  Qed.

  Lemma fold_equal: 
    (Equal s s') -> (eqA (fold f s i) (fold f s' i)).
  Proof. 
     Pattern s; Apply set_induction; Intros.
     Apply (Seq_trans ?? st) with i; Auto.
     Apply fold_1; Auto.
     Apply Seq_sym; Auto; Apply fold_1; Auto.
     Apply cardinal_inv_1; Rewrite <- (Equal_cardinal H0); Auto with set.
     Apply (Seq_trans ?? st) with (f x0 (fold f s0 i)); Auto.
     Apply fold_2 with eqA:=eqA; Auto.
     Apply Seq_sym; Auto; Apply fold_2 with eqA := eqA; Auto.
     Generalize H2 H1; Unfold Add Equal; Intros; 
       Elim (H4 y); Elim (H3 y); Intuition.
  Qed.
   
  Lemma fold_add: 
    ~(In x s) -> (eqA (fold f (add x s) i) (f x (fold f s i))).
  Proof. 
    Intros; Apply fold_2 with eqA:=eqA; Auto.
  Qed.

  End Fold.

  (** properties of [Equal] *)

  Lemma equal_refl : (s:t)(Equal s s).
  Proof.
  Unfold Equal; Intuition.
  Qed.

  Lemma equal_sym : (s,s':t)(Equal s s') -> (Equal s' s).
  Proof.
  Unfold Equal; Ground.
  Qed.

  Lemma equal_trans : 
    (s,u,v:t)(Equal s u) -> (Equal u v) -> (Equal s v).
  Proof.
  Ground. 
  Qed.

  Hints Resolve equal_refl equal_trans : set.
  Hints Immediate equal_sym : set.

  (** properties of [Subset] *)

  Lemma subset_refl : (s:t)(Subset s s). 
  Proof.
  Unfold Subset; Intuition.
  Qed.

  Lemma subset_antisym : 
    (s,s':t)(Subset s s') -> (Subset s' s) -> (Equal s s').
  Proof.
  Unfold Subset Equal; Intuition.
  Qed.

  Lemma subset_trans : 
   (s,t_,u:t)(Subset s t_) -> (Subset t_ u) -> (Subset s u).
  Proof.
  Unfold Subset; Intuition.
  Qed.

  Lemma subset_equal : 
    (s,s':t)(Equal s s') -> (Subset s s').
  Proof.
  Unfold Subset Equal; Ground.
  Qed.

  Lemma subset_empty :
    (s:t)(Subset empty s).
  Proof.
  Unfold Subset; Intuition.
  Absurd (In a empty); Auto.
  Qed.

  Lemma subset_remove_3 :
    (s1,s2:t)(x:elt)
    (Subset s1 s2) -> (Subset (remove x s1) s2).
  Proof.
  Unfold Subset; Intuition.
  Apply (H a).
  Apply remove_3 with x; Auto.
  Save.

  Lemma subset_diff :
    (s1,s2,s3:t)
    (Subset s1 s3) -> (Subset (diff s1 s2) s3).
  Proof.
  Unfold Subset; Intuition.
  Apply (H a).
  Apply diff_1 with s2; Intuition.
  Save.

  Lemma subset_add_3 :
    (s1,s2:t)(x:elt)
    (In x s2) -> (Subset s1 s2) -> (Subset (add x s1) s2).
  Proof.
  Unfold Subset; Intuition.
  Case (ME.eq_dec x a); Intuition.
  Apply In_1 with x; Auto.
  Apply (H0 a).
  Apply add_3 with x; Auto.
  Save.

  Lemma subset_add_2 :
    (s1,s2:t)(x:elt)
    (Subset s1 s2) -> (Subset s1 (add x s2)).
  Proof.
  Unfold Subset; Intuition.
  Save.

  Lemma in_subset :
    (s1,s2:t)(x:elt)
    (In x s1) -> (Subset s1 s2) -> (In x s2).
  Proof.
  Unfold Subset; Intuition.
  Save.

  Hints Resolve subset_refl subset_equal subset_antisym 
                subset_trans subset_empty subset_remove_3 
                subset_diff subset_add_3 subset_add_2 in_subset 
                : set.

  (** properties of [empty] *)

  Lemma empty_is_empty_1 : (s:t)(Empty s) -> (Equal s empty).
  Proof.
  Unfold Empty Equal; Ground.
  Absurd (In a empty); Auto.
  Qed.

  Lemma empty_is_empty_2 : (s:t)(Equal s empty) -> (Empty s).
  Proof.
  Unfold Empty Equal; Ground.
  Intro; Absurd (In a empty); Ground.
  Qed.

  Hints Resolve empty_is_empty_1 empty_is_empty_2 : set.

  (** properties of [add] *)

  Tactic Definition Add_ y x s :=
    Generalize (!add_1 s x y) (!add_2 s x y) (!add_3 s x y).

  Lemma add_equal : 
    (s:t)(x:elt)(In x s) -> (Equal (add x s) s).
  Proof.
  Unfold Equal; Intuition.
  Elim (ME.eq_dec x a); Add_ a x s; Ground.
  EAuto.
  Qed.

  Hints Resolve add_equal :set.

  (** properties of [remove] *)

  Tactic Definition Remove_ a x s :=
    Generalize (!remove_1 s x a) (!remove_2 s x a) (!remove_3 s x a).

  Lemma remove_equal: 
    (s:t)(x:elt)~(In x s) -> (Equal (remove x s) s).
  Proof.
  Unfold Equal; Intuition.
  EAuto.
  Elim (ME.eq_dec x a); Remove_ a x s; Ground.
  Absurd (In x s); Auto.
  Apply In_1 with a; Auto.
  Qed.

  Hints Resolve remove_equal : set.

  (** properties of [add] and [remove] *)

  Lemma add_remove: 
    (s:t)(x:elt)(In x s) -> (Equal (add x (remove x s)) s).
  Proof.
  Unfold Equal; Intuition.
  Elim (ME.eq_dec x a); Intuition.
  Apply In_1 with x; Auto.
  Add_ a x '(remove x s); Remove_ a x s; Ground.
  Elim (ME.eq_dec x a); Intuition.
  Qed.

  Lemma remove_add: 
    (s:t)(x:elt)~(In x s) -> (Equal (remove x (add x s)) s).
  Proof.
  Unfold Equal; Intuition.
  Elim (ME.eq_dec x a); Intuition.
  Absurd (In a (remove x (add x s))); Auto.
  Remove_ a x '(add x s); Add_ a x s; Ground.
  Elim (ME.eq_dec x a); Intuition.
  Elim H; Apply In_1 with a; Auto.
  Qed.

  Hints Immediate add_remove remove_add :set.

  (** properties of [singleton] *)

  Tactic Definition Singleton a x :=
    Generalize (!singleton_1 x a) (!singleton_2 x a).

  Lemma singleton_equal_add : (x:elt)
    (Equal (singleton x) (add x empty)).
  Proof.
  Unfold Equal; Intros;
  Singleton a x; Intuition.
  Apply H0; Clear H H0.
  Add_ a x empty; Intuition.
  Elim (ME.eq_dec x a); Intuition.
  Absurd (In a empty); Auto.
  Qed.

  Lemma singleton_cardinal: (x:elt)(cardinal (singleton x))=(S O).
  Proof.
  Intros.
  Rewrite (Equal_cardinal (singleton_equal_add x)).
  Replace O with (cardinal empty); Auto with set.
  Apply cardinal_2 with x; Auto with set.
  Qed.

  Hints Resolve singleton_equal_add singleton_cardinal : set.

  (** properties of [union] *)

  Tactic Definition Union a s s' := 
    Generalize (!union_1 s s' a) (!union_2 s s' a) (!union_3 s s' a).

  Lemma union_sym :
    (s,s':t)(Equal (union s s') (union s' s)).
  Proof.
  Unfold Equal; Intuition; Union a s s'; Union a s' s; Intuition.
  Qed.

  Lemma union_subset_equal : 
    (s,s':t)(Subset s s') -> (Equal (union s s') s').
  Proof.
  Unfold Subset Equal; Intuition; Union a s s'; Intuition.
  Qed.

  Lemma union_equal_1 : 
    (s,s',s'':t)(Equal s s') ->
    (Equal (union s s'') (union s' s'')).
  Proof.
  Unfold Equal; Intuition; 
  Union a s s''; Union a s' s''; Ground.
  Qed.

  Lemma union_equal_2 : 
    (s,s',s'':t)(Equal s' s'') ->
    (Equal (union s s') (union s s'')).
  Proof.
  Unfold Equal; Intuition; Union a s s'; Union a s s''; Ground.
  Qed.

  Lemma union_assoc : 
    (s,s',s'':t)
    (Equal (union (union s s') s'') (union s (union s' s''))).
  Proof.
  Unfold Equal; Intuition; Union a '(union s s') s''; 
  Union a s s'; Union a s '(union s' s''); Union a s' s''; Ground.
  Qed.

  Hints Resolve union_subset_equal union_equal_1 union_equal_2 
                union_assoc : set.
  Hints Immediate union_sym.

  Lemma add_union_singleton: 
    (s:t)(x:elt)(Equal (add x s) (union (singleton x) s)).
  Proof.
  Unfold Equal; Intuition; Add_ a x s; Singleton a x; Ground.
  Elim (ME.eq_dec x a); Ground.
  Union a '(singleton x) s; Ground.
  Qed.

  Lemma union_add: 
    (s,s':t)(x:elt)
    (Equal (union (add x s) s') (add x (union s s'))).
  Proof.
  Unfold Equal; Intuition.
  Union a '(add x s) s'; Intuition.
  Generalize (Add_add s x); Unfold Add; Ground.
  Generalize (Add_add (union s s') x); Unfold Add; Ground.
  Generalize (H0 a); Clear H0; Ground.
  Union a s s'; Intuition.
  Qed.

  Lemma union_subset_1 :
    (s,s':t)(Subset s (union s s')).
  Proof.
  Unfold Subset; Intuition.
  Qed.

  Lemma union_subset_2:
    (s,s':t)(Subset s' (union s s')).
  Proof.
  Unfold Subset; Intuition.
  Qed.

  Lemma subset_union :
    (s,s',s'':t)(Subset s s'') -> (Subset s' s'') -> 
    (Subset (union s s') s'').
  Proof.
  Unfold Subset; Intuition.
  Union a s s'; Intuition.
  Qed.

  Hints Resolve add_union_singleton union_add union_subset_1
                union_subset_2 subset_union : set.

  (** properties of [inter] *)

  Tactic Definition Inter a s s' :=
    Generalize (!inter_1 s s' a) (!inter_2 s s' a) (!inter_3 s s' a).

  Lemma inter_sym:
   (s,s':t)(Equal (inter s s') (inter s' s)).
  Proof.
  Unfold Equal; Intros s s' a.
  Inter a s s'; Inter a s' s; Intuition.
  Qed.

  Lemma inter_subset_equal: 
    (s,s':t)(Subset s s') -> (Equal (inter s s') s).
  Proof.
  Unfold Equal; Intros.
  Inter a s s'; Intuition.
  Qed.

  Lemma inter_equal_1: 
    (s,s',s'':t)(Equal s s') ->
    (Equal (inter s s'') (inter s' s'')).
  Proof.
  Unfold Equal; Intuition.
  Inter a s s''; Ground.
  Inter a s' s''; Ground.
  Qed.

  Lemma inter_equal_2: 
    (s,s',s'':t)(Equal s' s'') -> 
    (Equal (inter s s') (inter s s'')).
  Proof.
  Unfold Equal; Intuition.
  Inter a s s'; Ground.
  Inter a s s''; Ground.
  Qed.

  Hints Immediate inter_sym : set.
  Hints Resolve inter_subset_equal inter_equal_1 inter_equal_2 : set.

  Lemma inter_assoc: 
    (s,s',s'':t)
    (Equal (inter (inter s s') s'') (inter s (inter s' s''))).
  Proof.
  Unfold Equal; Intuition.
  Inter a '(inter s s') s''; Inter a s s'; Ground.
  Inter a s '(inter s' s''); Inter a s' s''; Ground.
  Qed.

  Lemma union_inter_1: 
    (s,s',s'':t)
    (Equal (inter (union s s') s'') (union (inter s s'') (inter s' s''))).
  Proof.
  Unfold Equal; Intuition.
  Inter a '(union s s') s''; Union a s s'; Ground.
  Union a '(inter s s'') '(inter s' s''); 
  Inter a s s''; Inter a s' s''; Ground.
  Qed.

  Lemma union_inter_2:
    (s,s',s'':t)
    (Equal (union (inter s s') s'') (inter (union s s'') (union s' s''))).
  Proof.
  Unfold Equal; Intuition.
  Union a '(inter s s') s''; Inter a s s'; Ground.
  Inter a '(union s s'') '(union s' s''); 
  Union a s s''; Union a s' s''; Ground.
  Qed.

  Lemma inter_add_1: 
    (s,s':t)(x:elt)(In x s') -> 
    (Equal (inter (add x s) s') (add x (inter s s'))).
  Proof.
  Unfold Equal; Intuition.
  Inter a '(add x s) s'; Add_ a x s; Ground.
  Elim (ME.eq_dec x a); Auto.
  Generalize (!Add_add (inter s s') x); Unfold Add; Ground.
  Generalize (H1 a); Clear H1; Intuition.
  Apply inter_3; Auto.
  Apply In_1 with x; Auto.
  Apply inter_3; Auto.
  Apply add_2; Auto.
  Inter a s s'; Intuition.
  Inter a s s'; Intuition.
  Qed.

  Lemma inter_add_2: 
    (s,s':t)(x:elt)~(In x s') -> 
    (Equal (inter (add x s) s') (inter s s')).
  Proof.
  Unfold Equal; Intuition.
  Inter a '(add x s) s'; Add_ a x s; Ground.
  Elim (ME.eq_dec x a); Intuition.
  Absurd (In x s'); Auto.
  Apply In_1 with a; Auto.
  Inter a s s'; Intuition.
  Qed.

  Hints Resolve inter_assoc union_inter_1 union_inter_2
                inter_add_1 inter_add_2 : set.

  (** properties of [Subset] *)

  Lemma inter_subset_1:
    (s,s':t)(Subset (inter s s') s).
  Proof.
  Unfold Subset; Intuition.
  Inter a s s'; Intuition.
  Qed.

  Lemma inter_subset_2:
    (s,s':t)(Subset (inter s s') s').
  Proof.
  Unfold Subset; Intros; Inter a s s'; Intuition.
  Qed.

  Lemma inter_subset_3:
    (s,s',s'':t)(Subset s'' s) -> (Subset s'' s') -> 
    (Subset s'' (inter s s')).
  Proof.
  Unfold Subset; Intuition.
  Qed.

  Hints Resolve inter_subset_1 inter_subset_2 inter_subset_3 : set.

  (** Properties of [diff] *)

  Tactic Definition Diff a s s' :=
    Generalize (!diff_1 s s' a) (!diff_2 s s' a) (!diff_3 s s' a).

  Lemma diff_subset: 
    (s,s':t)(Subset (diff s s') s).
  Proof.
  Unfold Subset; Intuition.
  Diff a s s'; Intuition.
  Qed.

  Lemma diff_subset_equal:
   (s,s':t)(Subset s s') -> (Equal (diff s s') empty).
  Proof.
  Unfold Subset Equal; Intuition.
  Diff a s s'; Ground.
  Absurd (In a empty); Auto.
  Qed.

  Lemma remove_inter_singleton: 
    (s:t)(x:elt)(Equal (remove x s) (diff s (singleton x))).
  Proof.
  Unfold Equal; Intuition.
  Remove_ a x s; Intuition.
  Elim (ME.eq_dec x a); Intuition.
  Diff a s '(singleton x); Intuition.
  Qed.

  Lemma diff_inter_empty:
    (s,s':t)(Equal (inter (diff s s') (inter s s')) empty). 
  Proof.
  Unfold Equal; Intuition.
  Inter a '(diff s s') '(inter s s'); Intuition.
  Inter a s s'; Diff a s s'; Intuition.
  Absurd (In a empty); Auto.
  Qed.

  (* TOMOVE *)
  Lemma In_dec : (x:elt)(s:t){ (In x s) }+ { ~(In x s) }.
  Proof.
  Intros x s; Generalize (!mem_1 s x) (!mem_2 s x); Case (mem x s); Intuition.
  Save.

  Lemma diff_inter_all: 
    (s,s':t)(Equal (union (diff s s') (inter s s')) s).
  Proof.
  Unfold Equal; Intuition.
  Union a '(diff s s') '(inter s s'); Intuition.
  Diff a s s'; Intuition.
  Inter a s s'; Intuition.
  Elim (In_dec a s'); Auto.
  Qed.

  Hints Resolve diff_subset diff_subset_equal
                remove_inter_singleton diff_inter_empty
                diff_inter_all : set.

  Lemma fold_plus: 
    (s:t)(p:nat)(fold [_]S s p)=(fold [_]S s O)+p.
  Proof.
  Assert st := (gen_st nat).
  Assert fe: (compat_op E.eq (eq ?) [_:elt]S). Unfold compat_op; Auto. 
  Assert fp: (transpose (eq ?) [_:elt]S). Unfold transpose;Auto.
  Intros s p;Pattern s;Apply set_induction.
  Intros; Rewrite (fold_1 st p [_:elt]S H).
  Rewrite (fold_1 st O [_:elt]S H); Trivial.
  Intros.
  Assert (p:nat)(s1:t)(Add x s0 s1) -> (fold [_]S s1 p) = (S (fold [_]S s0 p)).
  Change S with ([_]S x). 
  Intros; Apply fold_2 with eqA:=(eq nat); Auto.
  Rewrite (H2 p); Auto.
  Rewrite (H2 O); Auto.
  Rewrite H.
  Simpl; Auto.
  Qed.

  (** properties of [cardinal] *)

  Lemma empty_inter_1 : (s,s':t)(Empty s) -> (Empty (inter s s')).
  Proof.
  Unfold Empty; Intuition; Inter a s s'; Ground.
  Qed.

  Lemma empty_inter_2 : (s,s':t)(Empty s') -> (Empty (inter s s')).
  Proof.
  Unfold Empty; Intuition; Inter a s s'; Ground.
  Qed.

  Lemma empty_union_1 : (s,s':t)(Empty s) -> (Equal (union s s') s').
  Proof.
  Unfold Equal Empty; Intuition; Union a s s'; Ground.
  Qed.

  Lemma empty_union_2 : (s,s':t)(Empty s) -> (Equal (union s' s) s').
  Proof.
  Unfold Equal Empty; Intuition; Union a s' s; Ground.
  Qed.

  Lemma empty_diff_1 : (s,s':t)(Empty s) -> (Empty (diff s s')). 
  Proof.
  Unfold Empty Equal; Intuition; Diff a s s'; Ground.
  Qed.

  Lemma empty_diff_2 : (s,s':t)(Empty s) -> (Equal (diff s' s) s').
  Proof.
  Unfold Empty Equal; Intuition; Diff a s' s; Ground.
  Qed.

  Hints Resolve empty_inter_1 empty_inter_2
                empty_union_1 empty_union_2 
                empty_diff_1  empty_diff_2 : set.

  Lemma union_Add : (s,s',s'':t)(x:elt)
    (Add x s s') -> (Add x (union s s'') (union s' s'')).
  Proof. 
  Unfold Add; Intuition. 
  Union y s' s''; Ground.   
  Union y s' s''; Ground. 
  Union y s s''; Ground. 
  Qed.

  Lemma inter_Add : (s,s',s'':t)(x:elt)
    (In x s'') -> (Add x s s') -> (Add x (inter s s'') (inter s' s'')).
  Proof. 
  Unfold Add; Intuition. 
  Inter y s' s''; Ground.   
  Inter y s' s''; Ground. 
  Apply H4; Auto. 
  Elim (H0 y); Intuition.
  Apply In_1 with x; Auto. 
  Inter y s s''; Ground. 
  Qed.  

  Lemma union_Equal : (s,s',s'':t)(x:elt)
    (In x s'') -> (Add x s s') -> (Equal (union s s'') (union s' s'')).
  Proof. 
  Unfold Add Equal; Intuition. 
  Union a s s''; Ground. 
  Union a s' s''; Ground. 
  Elim (H0 a); Intuition. 
  Union a s s''; Ground. 
  Apply H10; Apply In_1 with x; Auto. 
  Qed.  

  Lemma inter_Add_2 :(s,s',s'':t)(x:elt)
    ~(In x s'') -> (Add x s s') -> (Equal (inter s s'') (inter s' s'')).
  Proof.
  Unfold Add Equal; Intuition. 
  Inter a s s''; Ground. 
  Elim (H0 a); Clear H0; Inter a s' s''; Ground. 
  Absurd (In x s''); Auto. 
  Apply In_1 with a; Auto. 
  Qed.

  Lemma not_in_union : (x:elt)(s,s':t)
    ~(In x s) -> ~(In x s') -> ~(In x (union s s')). 
  Proof.
  Intuition. 
  Union x s s'; Ground. 
  Qed.  

  Hints Resolve union_Add inter_Add union_Equal inter_Add_2 
                not_in_union : set. 

  Lemma fold_commutes:
    (A:Set)
    (f:elt->A->A)(i:A)(compat_op E.eq (eq ?) f) -> (transpose (eq ?) f) -> (s:t)(x:elt)
    (fold f s (f x i)) = (f x (fold f s i)).
  Proof.
  Intro A.
  LetTac st := (gen_st A).
  Intros; Pattern s; Apply set_induction.
  Intros.
  Rewrite (fold_1 st i f H1).
  Rewrite (fold_1 st (f x i) f H1).
  Apply H; Auto. 
  Intros. 
  Apply (Seq_trans ?? st) with (f x0 (fold f s0 (f x i))).
  Rewrite (fold_2 st (f x i) H H0 H2 H3).
  Apply (Seq_trans ?? st) with (f x0 (f x (fold f s0 i))).
  Apply H; Auto.
  Rewrite H1; Auto. 
  Rewrite (fold_2 st i H H0 H2 H3).
  Rewrite H1; Auto. 
  Qed.

  Lemma fold_union_inter: 
    (A:Set)
    (f:elt->A->A)(i:A)(compat_op E.eq (eq ?) f) -> (transpose (eq ?) f) -> 
    (s,s':t)(fold f (union s s') (fold f (inter s s') i))
     = (fold f s (fold f s' i)).
  Proof.
  Intro A.
  LetTac st := (gen_st A).
  Intros;Pattern s;Apply set_induction.
  Intros; Rewrite (fold_1 st (fold f s' i) f H1).
  Assert H2:(Empty (inter s0 s')). Auto with set.
  Rewrite (fold_1 st i f H2).
  Assert H3:(Equal (union s0 s') s'). Auto with set.
  Rewrite (fold_equal st i H H0 H3); Auto.
  Intros.
  Assert H4:(Add x (union s0 s') (union s'0 s')). 
  Auto with set. 
  Case (In_dec x s'); Intro.
  (* In x s' *)
  Assert (Add x (inter s0 s') (inter s'0 s')).
  Auto with set.
  Assert ~(In x (inter s0 s')).
  Inter x s0 s'; Intuition.
  Rewrite (fold_2 st i H H0 H6 H5).
  Rewrite (fold_2 st (fold f s' i) H H0 H2 H3).
  Assert (Equal (union s'0 s') (union s0 s')).
  Apply equal_sym; EAuto with set. 
  Rewrite (fold_equal st (f x (fold f s0 ⋂ s' i)) H H0 H7).
  Rewrite fold_commutes; Auto.
  (* ~(In x s') *)
  Rewrite (fold_2 st (fold f s' i) H H0 H2 H3).
  Assert (Equal (inter s'0 s') (inter s0 s')). 
  Apply equal_sym; EAuto with set. 
  Rewrite (fold_equal st i H H0 H5).
  Assert (Add x (union s0 s') (union s'0 s')). 
  Auto with set. 
  Assert ~(In x (union s0 s')). 
  Auto with set. 
  Rewrite (fold_2 st (fold f (inter s0 s') i) H H0 H7 H6).
  Auto. 
  Qed.

  Lemma fold_diff_inter: 
    (A:Set)(f:elt->A->A)(i:A)(compat_op E.eq (eq A) f) -> (transpose (eq A) f) ->
    (s,s':t)(fold f (diff s s') (fold f (inter s s') i))=(fold f s i).
  Proof.
  Intros.
  Assert st := (gen_st A).
  Rewrite <- (fold_union_inter i H H0 (diff s s') (inter s s')).
  Rewrite (fold_equal st i H H0 (diff_inter_empty s s')).
  Rewrite (fold_empty st).
  Apply fold_equal with eqA:=(eq A);Auto.
  Apply diff_inter_all.
  Qed.

  Lemma diff_inter_cardinal: 
    (s,s':t)(cardinal (diff s s'))+(cardinal (inter s s'))=(cardinal s).
  Proof.
  Intro s; Pattern s; Apply set_induction; Intuition.
  Assert (Empty (diff s0 s')); Auto with set.  
  Rewrite (empty_cardinal_2 H0). 
  Assert (Empty (inter s0 s')); Auto with set.  
  Rewrite (empty_cardinal_2 H1).
  Rewrite (empty_cardinal_2 H); Auto. 
  Intros.
  Do 3 Rewrite cardinal_fold.
  Rewrite <- fold_plus.
  Apply fold_diff_inter; Auto.
  Qed.

  Lemma subset_cardinal: 
    (s,s':t)(Subset s s') -> (le (cardinal s) (cardinal s')).
  Proof.
  Intros.
  Rewrite <- (diff_inter_cardinal s' s).
  Rewrite (Equal_cardinal (inter_sym s' s)).
  Rewrite (Equal_cardinal (inter_subset_equal H)); Auto with arith.
  Qed.

  Lemma union_inter_cardinal: 
    (s,s':t)(cardinal (union s s'))+(cardinal (inter s s'))
     = (cardinal s)+(cardinal s').
  Proof.
  Intros.
  Do 4 Rewrite cardinal_fold.
  Do 2 Rewrite <- fold_plus.
  Apply fold_union_inter;Auto.
  Qed.

  Lemma union_cardinal :
    (s,s':t)(le (cardinal (union s s')) (plus (cardinal s) (cardinal s'))).
  Proof.
    Intros; Generalize (union_inter_cardinal s s'); Intros; Omega.
  Qed.

  Lemma add_cardinal_1: 
    (s:t)(x:elt)(In x s) -> (cardinal (add x s))=(cardinal s).
  Proof.
  Auto with set.  
  Qed.

  Lemma add_cardinal_2: 
    (s:t)(x:elt)~(In x s) -> (cardinal (add x s))=(S (cardinal s)).
  Proof.
  Intros.
  Do 2 Rewrite cardinal_fold.
  Change S with ([_]S x); Apply fold_add with eqA:=(eq nat); Auto.
  Qed.

  Hints Resolve subset_cardinal union_cardinal
                add_cardinal_1 add_cardinal_2.

End Properties.

